import json
import platform
import shutil
import subprocess
import time
import zipfile
from collections import OrderedDict
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from os import path
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Union

import ifaddr
import ingenialogger
from virtual_drive.core import VirtualDrive

from ingenialink.canopen.network import CAN_BAUDRATE, CAN_DEVICE, CanopenNetwork
from ingenialink.canopen.servo import CanopenServo
from ingenialink.dictionary import Interface
from ingenialink.enums.register import REG_ACCESS, REG_DTYPE
from ingenialink.enums.servo import SERVO_STATE
from ingenialink.eoe.network import EoENetwork
from ingenialink.ethercat.network import EthercatNetwork
from ingenialink.ethernet.network import EthernetNetwork
from ingenialink.exceptions import ILError
from ingenialink.network import NET_DEV_EVT, SlaveInfo
from ingenialink.servo import DictionaryFactory
from ingenialink.virtual.network import VirtualNetwork
from ingeniamotion.exceptions import IMException, IMRegisterWrongAccess

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController

from ingeniamotion.metaclass import DEFAULT_AXIS, DEFAULT_SERVO, MCMetaClass

RUNNING_ON_WINDOWS = platform.system() == "Windows"
FILE_EXT_SFU = ".sfu"
FILE_EXT_LFU = ".lfu"


class Communication(metaclass=MCMetaClass):
    """Communication."""

    FORCE_SYSTEM_BOOT_COCO_REGISTER = "DRV_BOOT_COCO_FORCE"
    FORCE_SYSTEM_BOOT_MOCO_REGISTER = "DRV_BOOT_MOCO_MODE"
    SYSTEM_RESET_MOCO_REGISTER = "DRV_BOOT_RESET"

    PASSWORD_FORCE_BOOT_COCO = 0x424F4F54
    PASSWORD_FORCE_BOOT_MOCO = 0x426F6F74
    PASSWORD_SYSTEM_RESET = 0x72657365

    ENSEMBLE_FIRMWARE_EXTENSION = ".zfu"
    ENSEMBLE_TEMP_FOLDER = "ensemble_temp"
    ENSEMBLE_SLAVE_REV_NUM_MASK = 0x1000000

    def __init__(self, motion_controller: "MotionController") -> None:
        self.mc = motion_controller
        self.logger = ingenialogger.get_logger(__name__)
        self.__virtual_drive: Optional[VirtualDrive] = None

    def connect_servo_eoe(
        self,
        ip: str,
        dict_path: Optional[str] = None,
        alias: str = DEFAULT_SERVO,
        port: int = 1061,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> None:
        """Connect to target servo by Ethernet over EtherCAT

        Args:
            ip : servo IP.
            dict_path : servo dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            port : servo port. ``1061`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Raises:
            TypeError: If the dict_path argument is missing.
            FileNotFoundError: If the dict file doesn't exist.
            ingenialink.exceptions.ILError: If the servo's IP or port is incorrect.
        """
        if not dict_path:
            raise TypeError("dict_path argument is missing")
        self.__servo_connect(
            ip,
            dict_path,
            alias,
            port,
            servo_status_listener=servo_status_listener,
            net_status_listener=net_status_listener,
            is_eoe=True,
        )

    def connect_servo_ethernet(
        self,
        ip: str,
        dict_path: str,
        alias: str = DEFAULT_SERVO,
        port: int = 1061,
        connection_timeout: int = 1,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> None:
        """Connect to target servo by Ethernet

        Args:
            ip : servo IP
            dict_path : servo dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            port : servo port. ``1061`` by default.
            connection_timeout: Timeout in seconds for connection.
                ``1`` seconds by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Raises:
            FileNotFoundError: If the dict file doesn't exist.
            ingenialink.exceptions.ILError: If the servo's IP or port is incorrect.
        """
        if not path.isfile(dict_path):
            raise FileNotFoundError(f"{dict_path} file does not exist!")
        self.__servo_connect(
            ip,
            dict_path,
            alias,
            port,
            connection_timeout,
            servo_status_listener=servo_status_listener,
            net_status_listener=net_status_listener,
        )

    def connect_servo_virtual(
        self,
        dict_path: Optional[str] = None,
        alias: str = DEFAULT_SERVO,
        port: int = 1061,
        connection_timeout: int = 1,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> None:
        """Connect to the virtual drive using an ethernet communication.

        Args:
            dict_path : servo dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            port : servo port. ``1061`` by default.
            connection_timeout: Timeout in seconds for connection.
                ``1`` seconds by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Raises:
            FileNotFoundError: If the dict file doesn't exist.
            ingenialink.exceptions.ILError: If the servo's IP or port is incorrect.
        """
        if dict_path is not None and not path.isfile(dict_path):
            raise FileNotFoundError(f"{dict_path} file does not exist!")

        if self.__virtual_drive is None:
            self.__virtual_drive = VirtualDrive(port, dictionary_path=dict_path)
            self.__virtual_drive.start()

        self.mc.net[alias] = VirtualNetwork()
        net = self.mc.net[alias]
        servo = net.connect_to_slave(
            self.__virtual_drive.dictionary_path,
            port,
            connection_timeout,
            servo_status_listener=servo_status_listener,
            net_status_listener=net_status_listener,
        )

        self.mc.servos[alias] = servo
        self.mc.servo_net[alias] = alias

    def __servo_connect(
        self,
        ip: str,
        dict_path: str,
        alias: str,
        port: int = 1061,
        connection_timeout: int = 1,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
        is_eoe: bool = False,
    ) -> None:
        if not path.isfile(dict_path):
            raise FileNotFoundError(f"{dict_path} file does not exist!")

        self.mc.net[alias] = EthernetNetwork()
        net = self.mc.net[alias]
        servo = net.connect_to_slave(
            ip,
            dict_path,
            port,
            connection_timeout,
            servo_status_listener=servo_status_listener,
            net_status_listener=net_status_listener,
            is_eoe=is_eoe,
        )

        self.mc.servos[alias] = servo
        self.mc.servo_net[alias] = alias

    def connect_servo_eoe_service(
        self,
        ifname: str,
        dict_path: str,
        ip: str = "192.168.3.22",
        slave: int = 1,
        port: int = 1061,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> None:
        """Connect to target servo by Ethernet over EtherCAT

        Args:
            ifname : interface name. It should have format
                ``\\Device\\NPF_[...]``.
            dict_path : servo dictionary path.
            ip : IP address to be assigned to the servo.
            slave : slave index. ``1`` by default.
            port : servo port. ``1061`` by default.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Raises:
            NotImplementedError: If this method is run in Linux.
            FileNotFoundError: If the dict file doesn't exist.
            ValueError: ip must be a subnetwork of 192.168.3.0/24
            ingenialink.exceptions.ILError: If the EoE service is not running
            ingenialink.exceptions.ILError: If the EoE service cannot be started on the network
                                            interface.
        """
        if not RUNNING_ON_WINDOWS:
            raise NotImplementedError("EoE service only works on windows.")
        if not path.isfile(dict_path):
            raise FileNotFoundError(f"{dict_path} file does not exist!")
        if ifname not in self.mc.net:
            self.mc.net[ifname] = EoENetwork(ifname)
        net = self.mc.net[ifname]
        servo = net.connect_to_slave(
            slave,
            ip,
            dict_path,
            port,
            servo_status_listener=servo_status_listener,
            net_status_listener=net_status_listener,
        )
        servo.slave = slave  # type: ignore [attr-defined]
        self.mc.servos[alias] = servo
        self.mc.servo_net[alias] = ifname

    def connect_servo_eoe_service_interface_ip(
        self,
        interface_ip: str,
        dict_path: str,
        ip: str = "192.168.3.22",
        slave: int = 1,
        port: int = 1061,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> None:
        """Connect to target servo by Ethernet over EtherCAT

        Args:
            interface_ip : IP of the interface to be connected to.
            dict_path : servo dictionary path.
            ip : IP address to be assigned to the servo.
            slave : slave index. ``1`` by default.
            port : servo port. ``1061`` by default.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Raises:
            TypeError: If the dict_path argument is missing.
            IndexError: If interface index is out of range.
            FileNotFoundError: If the dict file doesn't exist.
            ValueError: ip must be a subnetwork of 192.168.3.0/24
            ingenialink.exceptions.ILError: If the EoE service is not running
            ingenialink.exceptions.ILError: If the EoE service cannot be started on the network
                                            interface.
        """
        self.connect_servo_eoe_service(
            self.get_ifname_from_interface_ip(interface_ip),
            dict_path,
            ip,
            slave,
            port,
            alias,
            servo_status_listener,
            net_status_listener,
        )

    def connect_servo_comkit(
        self,
        ip: str,
        coco_dict_path: str,
        moco_dict_path: str,
        alias: str = DEFAULT_SERVO,
        port: int = 1061,
        connection_timeout: int = 1,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> None:
        """Connect to target servo using a COM-KIT

        Args:
            ip : servo IP
            coco_dict_path : COCO dictionary path.
            moco_dict_path : MOCO dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            port : servo port. ``1061`` by default.
            connection_timeout: Timeout in seconds for connection.
                ``1`` seconds by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Raises:
            FileNotFoundError: If a dict file doesn't exist.
            ingenialink.exceptions.ILError: If the servo's IP or port is incorrect.
        """
        for dict_path in [coco_dict_path, moco_dict_path]:
            if not path.isfile(dict_path):
                raise FileNotFoundError(f"{dict_path} file does not exist!")
        self.__servo_connect(
            ip,
            moco_dict_path,
            alias,
            port,
            connection_timeout,
            servo_status_listener=servo_status_listener,
            net_status_listener=net_status_listener,
        )
        coco_dict = DictionaryFactory.create_dictionary(coco_dict_path, Interface.ETH)
        self.mc.servos[alias].dictionary += coco_dict

    @staticmethod
    def __get_adapter_name(index: int) -> str:
        """Returns the adapter name of an adapter based on its index.

        Args:
            index : position of interface selected in
                :func:`get_interface_name_list`.
        """
        adapter = ifaddr.get_adapters()[index]
        if RUNNING_ON_WINDOWS:
            return f"\\Device\\NPF_{bytes.decode(adapter.name)}"
        else:
            return str(adapter.name)

    def get_ifname_from_interface_ip(self, address: str) -> str:
        """Returns interface name based on the address ip of an interface.

        Args:
            address : ip expected adapter is expected to
            be configured with.

        Raises:
            ValueError: In case the input is not valid or the adapter
            is not found.

        Returns:
            Ifname of the controller.
        """
        try:
            index = self._get_interface_index_by_address(address)
        except IndexError:
            raise ValueError(
                f"Could not found a adapter configured as {address} "
                f"to connect as EtherCAT master"
            )
        return self.__get_adapter_name(index)

    @staticmethod
    def _get_interface_index_by_address(address: str) -> int:
        """Get the interface index by searching its IP address.

        Args:
            address: IP address of the adapter.

        Returns:
            Interface's index.

        Raises:
            IndexError: If no adapter matches the IP address.

        """
        for idx, adapter in enumerate(ifaddr.get_adapters()):
            for ip in adapter.ips:
                if ip.is_IPv4 and ip.ip == address:
                    return idx
        raise IndexError

    def get_ifname_by_index(self, index: int) -> str:
        """Return interface name by index.

        Args:
            index : position of interface selected in
                :func:`get_interface_name_list`.

        Returns:
            Real name of selected interface.
            It can be used for functions :func:`connect_servo_eoe_service` and
            :func:`connect_servo_ethercat`.

        Raises:
            IndexError: If interface index is out of range.

        """
        return self.__get_adapter_name(index)

    @staticmethod
    def get_interface_name_list() -> List[str]:
        """Get interface list.

        Returns:
            List with interface readable names.

        """
        return [x.nice_name for x in ifaddr.get_adapters()]

    def connect_servo_eoe_service_interface_index(
        self,
        if_index: int,
        dict_path: str,
        ip: str = "192.168.3.22",
        slave: int = 1,
        port: int = 1061,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> None:
        """Connect to target servo by Ethernet over EtherCAT

        Args:
            if_index : interface index in list given by function
                :func:`get_interface_name_list`.
            dict_path : servo dictionary path.
            ip : IP address to be assigned to the servo.
            slave : slave index. ``1`` by default.
            port : servo port. ``1061`` by default.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Raises:
            TypeError: If the dict_path argument is missing.
            IndexError: If interface index is out of range.
            FileNotFoundError: If the dict file doesn't exist.
            ValueError: ip must be a subnetwork of 192.168.3.0/24
            ingenialink.exceptions.ILError: If the EoE service is not running
            ingenialink.exceptions.ILError: If the EoE service cannot be started on the network
                                            interface.
        """
        self.connect_servo_eoe_service(
            self.get_ifname_by_index(if_index),
            dict_path,
            ip,
            slave,
            port,
            alias,
            servo_status_listener,
            net_status_listener,
        )

    def scan_servos_eoe_service(self, ifname: str) -> List[int]:
        """Return a List of available servos.

        Args:
            ifname : interface name. It should have format
                ``\\Device\\NPF_[...]``.
        Returns:
            Drives available in the target interface.

        Raises:
            NotImplementedError: If this method is run in Linux.
            ingenialink.exceptions.ILError: If the EoE service is not running
            TypeError: If some parameter has a wrong type.
        """
        if not RUNNING_ON_WINDOWS:
            raise NotImplementedError("EoE service only works on windows.")
        net = self.mc.net[ifname] if ifname in self.mc.net else EoENetwork(ifname)
        slaves = net.scan_slaves()
        if not isinstance(slaves, List):
            raise TypeError("Slaves are not saved in a list")
        return slaves

    def scan_servos_eoe_service_interface_index(self, if_index: int) -> List[int]:
        """Return a list of available servos.

        Args:
            if_index : interface index in list given by function
                :func:`get_interface_name_list`.
        Returns:
            Drives available in the target interface.

        Raises:
            IndexError: If interface index is out of range.
            ingenialink.exceptions.ILError: If the EoE service is not running

        """
        return self.scan_servos_eoe_service(self.get_ifname_by_index(if_index))

    def connect_servo_canopen(
        self,
        can_device: CAN_DEVICE,
        dict_path: str,
        node_id: int,
        baudrate: CAN_BAUDRATE = CAN_BAUDRATE.Baudrate_1M,
        channel: int = 0,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> None:
        """Connect to target servo by CANOpen.

        Args:
            can_device : CANOpen device type.
            dict_path : servo dictionary path.
            node_id : node id. It's posible scan node ids with
                :func:`scan_servos_canopen`.
            baudrate : communication baudrate. 1 Mbit/s by default.
            channel : CANopen device channel. ``0`` by default.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Raises:
            FileNotFoundError: If either of the dict files doesn't exist.
            ingenialink.exceptions.ILError: If CANOpen device type, node id or channel is incorrect.

        """

        if not path.isfile(dict_path):
            raise FileNotFoundError(f"Dict file {dict_path} does not exist!")

        net_key = f"{can_device}_{channel}_{baudrate}"
        if net_key not in self.mc.net:
            self.mc.net[net_key] = CanopenNetwork(can_device, channel, baudrate)
        net = self.mc.net[net_key]

        servo = net.connect_to_slave(node_id, dict_path, servo_status_listener, net_status_listener)
        self.mc.servos[alias] = servo
        self.mc.servo_net[alias] = net_key

    def connect_servo_ethercat(
        self,
        interface_name: str,
        slave_id: int,
        dict_path: str,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> None:
        """Connect to an EtherCAT slave.

        Args:
            interface_name : interface name. It should have format
                ``\\Device\\NPF_[...]``.
            slave_id: EtherCAT slave ID.
            dict_path : servo dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Raises:
            FileNotFoundError: If the dict file doesn't exist.

        """

        if not path.isfile(dict_path):
            raise FileNotFoundError(f"Dict file {dict_path} does not exist!")
        if interface_name not in self.mc.net:
            self.mc.net[interface_name] = EthercatNetwork(interface_name)
        net = self.mc.net[interface_name]
        servo = net.connect_to_slave(
            slave_id,
            dict_path,
            servo_status_listener=servo_status_listener,
            net_status_listener=net_status_listener,
        )
        self.mc.servos[alias] = servo
        self.mc.servo_net[alias] = interface_name

    def connect_servo_ethercat_interface_index(
        self,
        if_index: int,
        slave_id: int,
        dict_path: str,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> None:
        """Connect to an EtherCAT slave.

        Args:
            if_index : interface index in list given by function
                :func:`get_interface_name_list`.
            slave_id: EtherCAT slave ID.
            dict_path : servo dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        Raises:
            IndexError: If interface index is out of range.

        """
        self.connect_servo_ethercat(
            self.get_ifname_by_index(if_index),
            slave_id,
            dict_path,
            alias,
            servo_status_listener,
            net_status_listener,
        )

    def connect_servo_ethercat_interface_ip(
        self,
        interface_ip: str,
        slave_id: int,
        dict_path: str,
        alias: str = DEFAULT_SERVO,
        servo_status_listener: bool = False,
        net_status_listener: bool = False,
    ) -> None:
        """Connect to an EtherCAT slave.

        Args:
            interface_ip : IP of the interface to be connected to.
            slave_id: EtherCAT slave ID.
            dict_path : servo dictionary path.
            alias : servo alias to reference it. ``default`` by default.
            servo_status_listener : Toggle the listener of the servo for
                its status, errors, faults, etc.
            net_status_listener : Toggle the listener of the network
                status, connection and disconnection.

        """
        self.connect_servo_ethercat(
            self.get_ifname_from_interface_ip(interface_ip),
            slave_id,
            dict_path,
            alias,
            servo_status_listener,
            net_status_listener,
        )

    @staticmethod
    def scan_servos_ethercat_with_info(
        interface_name: str,
    ) -> OrderedDict[int, SlaveInfo]:
        """Scan a network adapter to get all connected EtherCAT slaves including slave information.

        Args:
            interface_name : interface name. It should have format
                ``\\Device\\NPF_[...]``.

        Returns:
            Dictionary of nodes available in the network and slave information.

        Raises:
            TypeError: If some parameter has a wrong type.
        """
        net = EthercatNetwork(interface_name)
        slaves_info = net.scan_slaves_info()
        return slaves_info

    def scan_servos_ethercat(
        self,
        interface_name: str,
    ) -> List[int]:
        """Scan a network adapter to get all connected EtherCAT slaves.

        Args:
            interface_name : interface name. It should have format
                ``\\Device\\NPF_[...]``.
        Returns:
            List of EtherCAT slaves available in the network.

        Raises:
            TypeError: If some parameter has a wrong type.
        """
        net = EthercatNetwork(interface_name)
        slaves = net.scan_slaves()
        return slaves

    def scan_servos_ethercat_interface_ip(self, interface_ip: str) -> List[int]:
        """Scan a network adapter to get all connected EtherCAT slaves.

        Args:
            interface_ip : IP of the interface to be connected to.

        Returns:
            List of EtherCAT slaves available in the network.

        """
        return self.scan_servos_ethercat(self.get_ifname_from_interface_ip(interface_ip))

    def scan_servos_ethercat_interface_index(self, if_index: int) -> List[int]:
        """Scan a network adapter to get all connected EtherCAT slaves.

        Args:
            if_index : interface index in list given by function
                :func:`get_interface_name_list`.
        Returns:
            List of EtherCAT slaves available in the network.

        Raises:
            IndexError: If interface index is out of range.

        """
        return self.scan_servos_ethercat(self.get_ifname_by_index(if_index))

    def scan_servos_canopen_with_info(
        self,
        can_device: CAN_DEVICE,
        baudrate: CAN_BAUDRATE = CAN_BAUDRATE.Baudrate_1M,
        channel: int = 0,
    ) -> OrderedDict[int, SlaveInfo]:
        """Scan CANOpen device network to get all nodes including slave information.

        Args:
            can_device : CANOpen device type.
            baudrate : communication baudrate. 1 Mbit/s by default.
            channel : CANOpen device channel. ``0`` by default.

        Returns:
            Dictionary of nodes available in the network and slave information.

        Raise:
            TypeError: If some parameter has a wrong type.
        """
        net_key = f"{can_device}_{channel}_{baudrate}"
        if net_key not in self.mc.net:
            self.mc.net[net_key] = CanopenNetwork(can_device, channel, baudrate)
        net = self.mc.net[net_key]

        if net is None:
            self.logger.warning(
                "Could not find any nodes in the network."
                "Device: %s, channel: %s and baudrate: %s.",
                can_device,
                channel,
                baudrate,
            )
            return []
        slaves_info = net.scan_slaves_info()
        return slaves_info

    def scan_servos_canopen(
        self,
        can_device: CAN_DEVICE,
        baudrate: CAN_BAUDRATE = CAN_BAUDRATE.Baudrate_1M,
        channel: int = 0,
    ) -> List[int]:
        """Scan CANOpen device network to get all nodes.

        Args:
            can_device : CANOpen device type.
            baudrate : communication baudrate. 1 Mbit/s by default.
            channel : CANOpen device channel. ``0`` by default.

        Returns:
            List of node ids available in the network.

        Raises:
            TypeError: If some parameter has a wrong type.
        """
        net_key = f"{can_device}_{channel}_{baudrate}"
        if net_key not in self.mc.net:
            self.mc.net[net_key] = CanopenNetwork(can_device, channel, baudrate)
        net = self.mc.net[net_key]

        if net is None:
            self.logger.warning(
                "Could not find any nodes in the network."
                "Device: %s, channel: %s and baudrate: %s.",
                can_device,
                channel,
                baudrate,
            )
            return []
        slaves = net.scan_slaves()
        return slaves

    def disconnect(self, servo: str = DEFAULT_SERVO) -> None:
        """Disconnect servo.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)
        network = self.mc._get_network(servo)
        network.disconnect_from_slave(drive)
        if isinstance(network, VirtualNetwork) and self.__virtual_drive:
            self.__virtual_drive.stop()
            self.__virtual_drive = None
        del self.mc.servos[servo]
        net_name = self.mc.servo_net.pop(servo)
        servo_count = list(self.mc.servo_net.values()).count(net_name)
        if servo_count == 0:
            del self.mc.net[net_name]

    def get_register(
        self, register: str, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> Union[int, float, str]:
        """Return the value of a target register.

        Args:
            register : register UID.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            Current register value.

        Raises:
            ingenialink.exceptions.ILAccessError: If the register access is write-only.
            IMRegisterNotExist: If the register doesn't exist.
            TypeError: If some parameter has a wrong type.

        """
        drive = self.mc.servos[servo]
        register_dtype = self.mc.info.register_type(register, axis, servo=servo)
        value = drive.read(register, subnode=axis)
        if register_dtype.value <= REG_DTYPE.S64.value and isinstance(value, int):
            return int(value)
        if not isinstance(value, (int, float, str)):
            raise TypeError("Register value is not a correct type of value.")
        return value

    def set_register(
        self,
        register: str,
        value: Union[int, float, str],
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        """Set a value of a target register.

        Args:
            register : register UID.
            value : new value for the register.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            TypeError: If the value is of the wrong type.
            IMRegisterNotExist: If the register doesn't exist.
            IMRegisterWrongAccess: If the register access is read-only.

        """
        drive = self.mc.servos[servo]
        register_dtype_value = self.mc.info.register_type(register, axis, servo=servo)
        register_access_type = self.mc.info.register_info(register, axis, servo=servo).access
        signed_int = [REG_DTYPE.S8, REG_DTYPE.S16, REG_DTYPE.S32, REG_DTYPE.S64]
        unsigned_int = [REG_DTYPE.U8, REG_DTYPE.U16, REG_DTYPE.U32, REG_DTYPE.U64]
        if register_dtype_value == REG_DTYPE.FLOAT and not isinstance(value, (int, float)):
            raise TypeError("Value must be a float")
        if register_dtype_value == REG_DTYPE.STR and not isinstance(value, str):
            raise TypeError("Value must be a string")
        if register_dtype_value in signed_int and not isinstance(value, int):
            raise TypeError("Value must be an int")
        if register_dtype_value in unsigned_int and (not isinstance(value, int) or value < 0):
            raise TypeError("Value must be an unsigned int")
        if register_access_type == REG_ACCESS.RO:
            raise IMRegisterWrongAccess(
                f"Register: {register} cannot write to a read-only register"
            )
        drive.write(register, value, subnode=axis)

    def subscribe_net_status(
        self, callback: Callable[[NET_DEV_EVT], None], servo: str = DEFAULT_SERVO
    ) -> None:
        """Add a callback to net status change event.

        Args:
            callback : when net status changes callback is called.
            servo : servo alias to reference it. ``default`` by default.

        """
        network = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        network.subscribe_to_status(drive.target, callback)

    def unsubscribe_net_status(
        self, callback: Callable[[NET_DEV_EVT], None], servo: str = DEFAULT_SERVO
    ) -> None:
        """Remove net status change event callback.

        Args:
            callback : callback to remove.
            servo : servo alias to reference it. ``default`` by default.

        """
        network = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        network.unsubscribe_from_status(drive.target, callback)

    def subscribe_servo_status(
        self, callback: Callable[[SERVO_STATE, None, int], Any], servo: str = DEFAULT_SERVO
    ) -> None:
        """Add a callback to servo status change event.

        Args:
            callback : when servo status changes callback is called.
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)
        drive.subscribe_to_status(callback)

    def unsubscribe_servo_status(
        self, callback: Callable[[SERVO_STATE, None, int], Any], servo: str = DEFAULT_SERVO
    ) -> None:
        """Remove servo status change event callback.

        Args:
            callback : callback to remove.
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)
        drive.unsubscribe_from_status(callback)

    def load_firmware_canopen(
        self,
        fw_file: str,
        servo: str = DEFAULT_SERVO,
        status_callback: Optional[Callable[[str], None]] = None,
        progress_callback: Optional[Callable[[int], None]] = None,
        error_enabled_callback: Optional[Callable[[bool], None]] = None,
    ) -> None:
        """Load firmware via CANopen.

        Args:
            fw_file : Firmware file path.
            servo : servo alias to reference it. ``default`` by default.
            status_callback : callback with status.
            progress_callback : callback with progress.
            error_enabled_callback : callback with errors enabled.

        Raises:
            ValueError: If servo is not connected via CANopen.

        """
        net = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        if not isinstance(net, CanopenNetwork):
            raise ValueError("Target servo is not connected via CANopen")
        if not isinstance(drive, CanopenServo):
            raise ValueError("Target servo is not a Canopen servo")
        if status_callback is None:
            status_callback = partial(self.logger.info, "Load firmware status: %s")
        if progress_callback is None:
            progress_callback = partial(self.logger.info, "Load firmware progress: %s")
        if fw_file.endswith(self.ENSEMBLE_FIRMWARE_EXTENSION):
            self.__load_ensemble_fw_canopen(
                net,
                fw_file,
                drive,
                status_callback,
                progress_callback,
                error_enabled_callback,
            )
        else:
            net.load_firmware(
                int(drive.target),
                fw_file,
                status_callback,
                progress_callback,
                error_enabled_callback,
            )

    @staticmethod
    def __get_boot_in_app(fw_file: str) -> bool:
        """Return true if the bootloader is included in the application (file extension is .sfu).

        Args:
            fw_file: Path to the FW file.

        Returns:
            True if the bootloader is included in the application.

        Raises:
            ValueError: If the firmware file has the wrong extension.
        """
        if not fw_file.endswith((FILE_EXT_SFU, FILE_EXT_LFU)):
            raise ValueError(
                f"Firmware file should have extension {FILE_EXT_SFU} or {FILE_EXT_LFU}."
            )
        return fw_file.endswith(FILE_EXT_SFU)

    def load_firmware_ecat(
        self, ifname: str, fw_file: str, slave: int = 1, boot_in_app: Optional[bool] = None
    ) -> None:
        """Load firmware via ECAT.

        Args:
            ifname : interface name. It should have format
                ``\\Device\\NPF_[...]``.
            fw_file : Firmware file path.
            slave : slave index. ``1`` by default.
            boot_in_app: true if the bootloader is included in the application, false otherwise.
                If None, the file extension is used to define it.

        Raises:
            FileNotFoundError: If the firmware file cannot be found.
            ValueError: If the firmware file has the wrong extension.
            ingenialink.exceptions.ILFirmwareLoadError: If no slave is detected.
            ingenialink.exceptions.ILFirmwareLoadError: If the FoE write operation is not successful
            NotImplementedError: If FoE is not implemented for the current OS and architecture

        """

        net = EthercatNetwork(ifname)
        if fw_file.endswith(self.ENSEMBLE_FIRMWARE_EXTENSION):
            self.__load_ensemble_fw_ecat(net, fw_file, slave, boot_in_app)
        else:
            boot_in_app = self.__get_boot_in_app(fw_file) if boot_in_app is None else boot_in_app
            net.load_firmware(fw_file, boot_in_app, slave)

    def load_firmware_ecat_interface_index(
        self, if_index: int, fw_file: str, slave: int = 1
    ) -> None:
        """Load firmware via ECAT.

        Args:
            if_index : interface index in list given by function
                :func:`get_interface_name_list`.
            fw_file : Firmware file path.
            slave : slave index. ``1`` by default.

        Raises:
            IndexError: If interface index is out of range.
            FileNotFoundError: If the firmware file cannot be found.
            ingenialink.exceptions.ILFirmwareLoadError: If no slave is detected.
            ingenialink.exceptions.ILFirmwareLoadError: If the FoE write operation is not successful
            NotImplementedError: If FoE is not implemented for the current OS and architecture

        """
        self.load_firmware_ecat(self.get_ifname_by_index(if_index), fw_file, slave)

    def load_firmware_ethernet(
        self, ip: str, fw_file: str, ftp_user: Optional[str] = None, ftp_pwd: Optional[str] = None
    ) -> None:
        """Load firmware via Ethernet. Boot mode is needed to load firmware.

        .. warning::
            After functions ends, the servo will take a moment to load firmware.
            During the process, the servo will be not operative.

        Args:
            ip : servo IP.
            fw_file : Firmware file path.
            ftp_user : FTP user to connect with.
            ftp_pwd : FTP password for the given user.

        """
        net = EthernetNetwork()
        ftp_user = ftp_user or "Ingenia"
        ftp_pwd = ftp_pwd or "Ingenia"
        net.load_firmware(fw_file, ip, ftp_user, ftp_pwd)

    @staticmethod
    def __ftp_ping(ip: str) -> bool:
        command = ["ping", ip]
        return subprocess.call(command) == 0

    def boot_mode_and_load_firmware_ethernet(
        self,
        fw_file: str,
        servo: str = DEFAULT_SERVO,
        ftp_user: Optional[str] = None,
        ftp_pwd: Optional[str] = None,
    ) -> None:
        """Set servo to boot mode and load firmware. Servo is disconnected.

        .. warning::
            After functions ends, the servo will take a moment to load firmware.
            During the process, the servo will be not operative.

        Args:
            fw_file : Firmware file path.
            servo : servo alias to reference it. ``default`` by default.
            ftp_user : FTP user to connect with.
            ftp_pwd : FTP password for the given user.

        Raises:
            ValueError: If servo is not connected via Ethernet.

        """
        net = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        ip = str(drive.target)
        if not isinstance(net, EthernetNetwork):
            raise ValueError("Target servo is not connected via Ethernet")
        self.boot_mode(servo)
        timeout = 5
        init_time = time.time()
        ftp_ready = False
        while not ftp_ready and time.time() - init_time < timeout:
            ftp_ready = self.__ftp_ping(ip)
            time.sleep(1)
        self.load_firmware_ethernet(ip, fw_file, ftp_user, ftp_pwd)

    def boot_mode(self, servo: str = DEFAULT_SERVO) -> None:
        """Set servo to boot mode. Servo is disconnected.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        """
        net = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        net.stop_status_listener()
        drive.stop_status_listener()
        try:
            self.mc.communication.set_register(
                self.FORCE_SYSTEM_BOOT_COCO_REGISTER,
                self.PASSWORD_FORCE_BOOT_COCO,
                servo=servo,
                axis=0,
            )
        except ILError:
            pass
        self.disconnect(servo)

    def load_firmware_moco(
        self,
        fw_file: str,
        servo: str = DEFAULT_SERVO,
    ) -> None:
        """Load firmware to the Motion Core.

        .. warning::
            After functions ends, the servo will take a moment to load firmware.
            During the process, the servo will be not operative.

        Args:
            fw_file : Firmware file path.
            servo : servo alias to reference it. ``default`` by default.

        Raises:
            ValueError: If servo is not connected via Ethernet.

        """
        default_node = 10
        default_subnode = 1
        default_port = 1061
        net = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        ip = str(drive.target)
        if not isinstance(net, EthernetNetwork):
            raise ValueError("Target servo is not connected via Ethernet")
        net.load_firmware_moco(default_node, default_subnode, ip, default_port, fw_file)

    def boot_mode_moco(self, servo: str = DEFAULT_SERVO) -> None:
        """Set the Motion Core to boot mode.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        """
        net = self.mc._get_network(servo)
        drive = self.mc._get_drive(servo)
        net.stop_status_listener()
        drive.stop_status_listener()
        try:
            self.mc.communication.set_register(
                self.FORCE_SYSTEM_BOOT_MOCO_REGISTER,
                self.PASSWORD_FORCE_BOOT_MOCO,
                servo=servo,
            )
        except ILError as e:
            self.logger.debug(e)
        try:
            self.mc.communication.set_register(
                self.SYSTEM_RESET_MOCO_REGISTER,
                self.PASSWORD_FORCE_BOOT_MOCO,
                servo=servo,
            )
        except ILError as e:
            self.logger.debug(e)

    def __load_ensemble_fw_canopen(
        self,
        net: CanopenNetwork,
        fw_file: str,
        slave: CanopenServo,
        status_callback: Optional[Callable[[str], None]] = None,
        progress_callback: Optional[Callable[[int], None]] = None,
        error_enabled_callback: Optional[Callable[[bool], None]] = None,
    ) -> None:
        """Load FW to an ensemble of servos through Canopen (in parallel).

        The FW files that are contained in the fw_file will be loaded by selecting any slave which
        is part of the ensemble.

        Args:
            net: Canopen network.
            fw_file: Path to the ensemble FW file.
            slave: Servo object.

        Raises:
            IMException: If the load FW process of any slave failed.
        """
        mapping = self.__unzip_ensemble_fw_file(fw_file)
        scanned_slaves = net.scan_slaves_info()
        first_slave_in_ensemble = self.__check_ensemble(scanned_slaves, int(slave.target), mapping)
        dictionary_path = slave.dictionary.path
        connected_drives = {int(slave.target): slave for slave in net.servos}
        for slave_id_offset in mapping:
            slave_id = first_slave_in_ensemble + slave_id_offset
            if slave_id not in connected_drives:
                net.connect_to_slave(slave_id, dictionary_path)
        thread_results = {}
        with ThreadPoolExecutor() as executor:
            for slave_id_offset, fw_file_prod_code in mapping.items():
                slave_id = first_slave_in_ensemble + slave_id_offset
                thread_results[slave_id] = executor.submit(
                    net.load_firmware,
                    slave_id,
                    fw_file_prod_code[0],
                    status_callback,
                    progress_callback,
                    error_enabled_callback,
                )

        exception = None
        for slave_id in thread_results:
            if slave_id not in connected_drives:
                net.disconnect_from_slave(connected_drives[slave_id])
            exception = thread_results[slave_id].exception()
            break

        shutil.rmtree(self.ENSEMBLE_TEMP_FOLDER)

        if exception is not None:
            raise IMException(
                f"Load of FW in slave {slave_id} of ensemble failed. Exception: {exception}"
            )

    def __load_ensemble_fw_ecat(
        self, net: EthercatNetwork, fw_file: str, slave: int, boot_in_app: Optional[bool]
    ) -> None:
        """Load FW to an ensemble of servos through Ethercat.

        The FW files that are contained in the fw_file will be loaded by selecting any slave which
        is part of the ensemble.

        Args:
            net: Ethercat network.
            fw_file: Path to the ensemble FW file.
            slave: Slave ID (any slave in the ensemble)
            boot_in_app: true if the bootloader is included in the application, false otherwise.
                If None, the file extension is used to define it.
        """
        mapping = self.__unzip_ensemble_fw_file(fw_file)
        scanned_slaves = net.scan_slaves_info()
        first_slave_in_ensemble = self.__check_ensemble(scanned_slaves, slave, mapping)
        try:
            for slave_id_offset, fw_file_prod_code in mapping.items():
                boot_in_app_drive = (
                    self.__get_boot_in_app(fw_file_prod_code[0])
                    if boot_in_app is None
                    else boot_in_app
                )
                net.load_firmware(
                    fw_file_prod_code[0],
                    boot_in_app_drive,
                    first_slave_in_ensemble + slave_id_offset,
                )
        except ILError as e:
            raise e
        finally:
            shutil.rmtree(self.ENSEMBLE_TEMP_FOLDER)

    def __check_ensemble(
        self,
        scanned_slaves: OrderedDict[int, SlaveInfo],
        slave_id: int,
        mapping: dict[int, tuple[str, int, int]],
    ) -> int:
        """Check that a slave is part of the ensemble described in the mapping argument and the
        ensemble is complete (all drives described in the mapping are contained in the list of
        scanned slaves).

        It returns the ID of the first drive in the ensemble.

        Args:
            scanned_slaves: Scanned slaves with the respective information.
            slave_id: Slave ID to be checked.
            mapping: Mapping of the ensemble.

        Raises:
            IMException: If the slave ID is not in the scanned slaves list.
            IMException: If the ensemble described in the mapping can not be found in the list of
                scanned slaves.

        Returns:
            The ID of the first drive in the ensemble.
        """
        slave_id_offset = self.__check_slave_in_ensemble(scanned_slaves[slave_id], mapping)
        first_slave = slave_id - slave_id_offset
        for map_slave_id_offset in mapping:
            map_slave_id = first_slave + map_slave_id_offset
            if map_slave_id not in scanned_slaves:
                raise IMException(f"Wrong ensemble. The slave {map_slave_id - 1} is not detected.")
            map_slave_info = scanned_slaves[map_slave_id]
            if (
                map_slave_info.product_code != mapping[map_slave_id_offset][1]
                or map_slave_info.revision_number != mapping[map_slave_id_offset][2]
            ):
                raise IMException(
                    f"Wrong ensemble. The slave {map_slave_id} has wrong product code or revision number."
                )
        return first_slave

    def __check_slave_in_ensemble(
        self, slave_info: SlaveInfo, mapping: dict[int, tuple[str, int, int]]
    ) -> int:
        """Check that a slave is part of the ensemble described in the mapping argument.

        Returns the ID offset (relative position in the ensemble) of the selected slave.

        Args:
            slave_info: Info of the slave to be checked.
            mapping: Mapping of the ensemble.

        Raises:
            IMException: If the slave is not part of the ensemble.

        Returns:
            The ID offset (relative position in the ensemble) of the selected slave.
        """
        for slave_id_offset in mapping:
            _, product_code, revision_number = mapping[slave_id_offset]
            if product_code == slave_info.product_code and (
                revision_number & self.ENSEMBLE_SLAVE_REV_NUM_MASK
            ) == (slave_info.revision_number & self.ENSEMBLE_SLAVE_REV_NUM_MASK):
                return slave_id_offset
        raise IMException("The selected drive is not part of the ensemble.")

    def __unzip_ensemble_fw_file(self, fw_file: str) -> dict[int, tuple[str, int, int]]:
        """Unzip the ensemble FW file and return the mapping.

        Args:
            fw_file: Ensemble FW file to be unzipped.

        Raises:
            IMException: If the file extension is incorrect.

        Returns:
            Mapping described in the ensemble FW file.
                Dict{slave_id_offset: (fw_file, product_code, revision_number)}
        """
        with zipfile.ZipFile(fw_file, "r") as zip_ref:
            zip_ref.extractall(self.ENSEMBLE_TEMP_FOLDER)
        with open(path.join(self.ENSEMBLE_TEMP_FOLDER, "mapping.json")) as f:
            mapping_info = json.load(f)
        mapping = {}
        for drive in mapping_info["drives"]:
            slave_id_offset = drive["slave_id_offset"]
            fw_file = path.join(path.abspath(self.ENSEMBLE_TEMP_FOLDER), drive["fw_file"])
            product_code = drive["product_code"]
            revision_number = drive["revision_number"]
            mapping[slave_id_offset] = fw_file, product_code, revision_number

        return mapping



import pytest
from ingenialink.exceptions import ILError

USER_UNDER_VOLTAGE_ERROR_OPTION_CODE_REGISTER = "ERROR_PROT_UNDER_VOLT_OPTION"
USER_UNDER_VOLTAGE_LEVEL_REGISTER = "DRV_PROT_USER_UNDER_VOLT"


@pytest.fixture
def error_number(motion_controller):
    mc, alias = motion_controller
    return mc.errors.get_number_total_errors(servo=alias)


@pytest.fixture
def generate_drive_errors(motion_controller):
    mc, alias = motion_controller
    errors_list = [
        {"code": 0x3241, "register": "DRV_PROT_USER_UNDER_VOLT", "value": 100},
        {"code": 0x4303, "register": "DRV_PROT_USER_OVER_TEMP", "value": 1},
        {"code": 0x3231, "register": "DRV_PROT_USER_OVER_VOLT", "value": 1},
        {"code": 0x4304, "register": "DRV_PROT_USER_UNDER_TEMP", "value": 200},
    ]
    error_code_list = []
    for item in errors_list:
        mc.motion.fault_reset(servo=alias)
        old_value = mc.communication.get_register(item["register"], servo=alias)
        mc.communication.set_register(item["register"], item["value"], servo=alias)
        try:
            mc.motion.motor_enable(servo=alias)
        except ILError:
            pass
        error_code_list.append(item["code"])
        try:
            mc.communication.set_register(item["register"], old_value, servo=alias)
        except ILError:
            # Sometimes fails with EVE-XCR-E
            mc.communication.set_register(item["register"], old_value, servo=alias)
    yield error_code_list[::-1]
    mc.motion.fault_reset(servo=alias)


@pytest.fixture
def force_warning(motion_controller):
    mc, alias = motion_controller
    mc.communication.set_register(USER_UNDER_VOLTAGE_ERROR_OPTION_CODE_REGISTER, 1, servo=alias)
    mc.communication.set_register(USER_UNDER_VOLTAGE_LEVEL_REGISTER, 100, servo=alias)
    mc.motion.motor_enable(servo=alias)
    yield
    mc.communication.set_register(USER_UNDER_VOLTAGE_ERROR_OPTION_CODE_REGISTER, 0, servo=alias)
    mc.communication.set_register(USER_UNDER_VOLTAGE_LEVEL_REGISTER, 10, servo=alias)


class TestErrors:
    @pytest.mark.smoke
    def test_get_last_error(self, motion_controller, generate_drive_errors):
        mc, alias = motion_controller
        last_error, subnode, warning = mc.errors.get_last_error(servo=alias)
        assert last_error == generate_drive_errors[0]
        mc.motion.fault_reset(servo=alias)
        last_error, subnode, warning = mc.errors.get_last_error(servo=alias)
        assert last_error == 0
        assert subnode is None
        assert warning is None

    @pytest.mark.smoke
    def test_get_last_buffer_error(self, motion_controller, generate_drive_errors):
        mc, alias = motion_controller
        last_error, subnode, warning = mc.errors.get_last_buffer_error(servo=alias)
        assert last_error == generate_drive_errors[0]

    @pytest.mark.smoke
    def test_get_buffer_error_by_index(self, motion_controller, generate_drive_errors):
        mc, alias = motion_controller
        index_list = [2, 1, 3, 0]
        for i in index_list:
            last_error, subnode, warning = mc.errors.get_buffer_error_by_index(i, servo=alias)
            assert last_error == generate_drive_errors[i]

    @pytest.mark.smoke
    def test_get_buffer_error_by_index_exception(self, motion_controller, generate_drive_errors):
        mc, alias = motion_controller
        with pytest.raises(ValueError):
            mc.errors.get_buffer_error_by_index(33, servo=alias)

    @pytest.mark.smoke
    def test_get_number_total_errors(self, motion_controller, error_number, generate_drive_errors):
        mc, alias = motion_controller
        test_error_number = mc.errors.get_number_total_errors(servo=alias)
        assert test_error_number == error_number + len(generate_drive_errors)

    @pytest.mark.smoke
    def test_get_all_errors(self, motion_controller, generate_drive_errors):
        mc, alias = motion_controller
        test_all_errors = mc.errors.get_all_errors(servo=alias)
        for i, code_error in enumerate(generate_drive_errors):
            test_code_error, axis, warning = test_all_errors[i]
            assert test_code_error == code_error

    @pytest.mark.smoke
    def test_is_fault_active(self, motion_controller, generate_drive_errors):
        mc, alias = motion_controller
        assert mc.errors.is_fault_active(servo=alias)
        mc.motion.fault_reset(servo=alias)
        assert not mc.errors.is_fault_active(servo=alias)

    @pytest.mark.smoke
    @pytest.mark.usefixtures("force_warning")
    def test_is_warning_active(self, motion_controller):
        mc, alias = motion_controller
        assert mc.errors.is_warning_active(servo=alias)
        mc.communication.set_register(USER_UNDER_VOLTAGE_LEVEL_REGISTER, 10, servo=alias)
        assert not mc.errors.is_warning_active(servo=alias)

    @pytest.mark.virtual
    @pytest.mark.smoke
    @pytest.mark.parametrize(
        "error_code, affected_module, error_type, error_msg",
        [
            (0x3241, "Power stage", "Cyclic", "User Under-voltage detected"),
            (0x4303, "Power stage", "Cyclic", "Over-temperature detected (user limit)"),
            (0x3231, "Power stage", "Cyclic", "User Over-voltage detected"),
            (0x4304, "Power stage", "Cyclic", "Under-temperature detected (user limit)"),
        ],
    )
    def test_get_error_data(
        self, motion_controller, error_code, affected_module, error_type, error_msg
    ):
        mc, alias = motion_controller
        test_id, test_aff_mod, test_type, test_msg = mc.errors.get_error_data(
            error_code, servo=alias
        )
        assert error_code == int(test_id, base=16)
        assert test_aff_mod == affected_module
        assert test_type == error_type
        assert test_msg == error_msg

    @pytest.mark.parametrize(
        "function",
        [
            "get_last_error",
            "get_buffer_error_by_index",
            "get_number_total_errors",
        ],
    )
    @pytest.mark.smoke
    @pytest.mark.virtual
    def test_wrong_type_exception(self, mocker, motion_controller, function):
        mc, alias = motion_controller
        mocker.patch.object(mc.communication, "get_register", return_value="invalid_value")
        with pytest.raises(TypeError):
            getattr(mc.errors, function)(servo=alias)



from enum import IntEnum
from os import path
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import ingenialogger
from ingenialink.exceptions import ILError
from ingenialink.canopen.network import CanopenNetwork, CAN_BAUDRATE
from ingenialink.ethernet.servo import EthernetServo

from ingeniamotion.homing import Homing
from ingeniamotion.metaclass import DEFAULT_AXIS, DEFAULT_SERVO, MCMetaClass
from ingeniamotion.exceptions import IMException
from ingeniamotion.feedbacks import Feedbacks
from ingeniamotion.enums import (
    GeneratorMode,
    PhasingMode,
    CommutationMode,
    FilterNumber,
    FilterSignal,
    FilterType,
)

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController


class TYPE_SUBNODES(IntEnum):
    COCO = 0
    MOCO = 1


class Configuration(Homing, Feedbacks, metaclass=MCMetaClass):
    """Configuration."""

    class BrakeOverride(IntEnum):
        """Brake override configuration enum"""

        OVERRIDE_DISABLED = 0
        RELEASE_BRAKE = 1
        ENABLE_BRAKE = 2

    BRAKE_OVERRIDE_REGISTER = "MOT_BRAKE_OVERRIDE"
    PROFILE_MAX_ACCELERATION_REGISTER = "PROF_MAX_ACC"
    PROFILE_MAX_DECELERATION_REGISTER = "PROF_MAX_DEC"
    PROFILE_MAX_VELOCITY_REGISTER = "PROF_MAX_VEL"
    MAX_VELOCITY_REGISTER = "CL_VEL_REF_MAX"
    POWER_STAGE_FREQUENCY_SELECTION_REGISTER = "DRV_PS_FREQ_SELECTION"
    POWER_STAGE_FREQUENCY_REGISTERS = [
        "DRV_PS_FREQ_1",
        "DRV_PS_FREQ_2",
        "DRV_PS_FREQ_3",
        "DRV_PS_FREQ_4",
    ]
    POSITION_AND_VELOCITY_LOOP_RATE_REGISTER = "DRV_POS_VEL_RATE"
    CURRENT_LOOP_RATE_REGISTER = "CL_CUR_FREQ"
    STATUS_WORD_REGISTER = "DRV_STATE_STATUS"
    PHASING_MODE_REGISTER = "COMMU_PHASING_MODE"
    GENERATOR_MODE_REGISTER = "FBK_GEN_MODE"
    MOTOR_POLE_PAIRS_REGISTER = "MOT_PAIR_POLES"
    STO_STATUS_REGISTER = "DRV_PROT_STO_STATUS"
    VELOCITY_LOOP_KP_REGISTER = "CL_VEL_PID_KP"
    VELOCITY_LOOP_KI_REGISTER = "CL_VEL_PID_KI"
    VELOCITY_LOOP_KD_REGISTER = "CL_VEL_PID_KD"
    POSITION_LOOP_KP_REGISTER = "CL_POS_PID_KP"
    POSITION_LOOP_KI_REGISTER = "CL_POS_PID_KI"
    POSITION_LOOP_KD_REGISTER = "CL_POS_PID_KD"
    RATED_CURRENT_REGISTER = "MOT_RATED_CURRENT"
    MAX_CURRENT_REGISTER = "CL_CUR_REF_MAX"
    COMMUTATION_MODE_REGISTER = "MOT_COMMU_MOD"
    BUS_VOLTAGE_REGISTER = "DRV_PROT_VBUS_VALUE"
    POSITION_TO_VELOCITY_RATIO_REGISTER = "PROF_POS_VEL_RATIO"
    FILTER_TYPE_REGISTER = "CL_{}_FILTER{}_TYPE"
    FILTER_FREQ_REGISTER = "CL_{}_FILTER{}_FREQ"
    FILTER_Q_REGISTER = "CL_{}_FILTER{}_Q"
    FILTER_GAIN_REGISTER = "CL_{}_FILTER{}_GAIN"

    STATUS_WORD_OPERATION_ENABLED_BIT = 0x04
    STATUS_WORD_COMMUTATION_FEEDBACK_ALIGNED_BIT = 0x4000
    STO1_ACTIVE_BIT = 0x1
    STO2_ACTIVE_BIT = 0x2
    STO_SUPPLY_FAULT_BIT = 0x4
    STO_ABNORMAL_FAULT_BIT = 0x8
    STO_REPORT_BIT = 0x10

    STO_ACTIVE_STATE = 4
    STO_INACTIVE_STATE = 23
    STO_LATCHED_STATE = 31

    PRODUCT_ID_REGISTERS = {
        TYPE_SUBNODES.COCO: "DRV_ID_PRODUCT_CODE_COCO",
        TYPE_SUBNODES.MOCO: "DRV_ID_PRODUCT_CODE",
    }
    REVISION_NUMBER_REGISTERS = {
        TYPE_SUBNODES.COCO: "DRV_ID_REVISION_NUMBER_COCO",
        TYPE_SUBNODES.MOCO: "DRV_ID_REVISION_NUMBER",
    }
    SERIAL_NUMBER_REGISTERS = {
        TYPE_SUBNODES.COCO: "DRV_ID_SERIAL_NUMBER_COCO",
        TYPE_SUBNODES.MOCO: "DRV_ID_SERIAL_NUMBER",
    }
    SOFTWARE_VERSION_REGISTERS = {
        TYPE_SUBNODES.COCO: "DRV_APP_COCO_VERSION",
        TYPE_SUBNODES.MOCO: "DRV_ID_SOFTWARE_VERSION",
    }
    VENDOR_ID_COCO_REGISTER = "DRV_ID_VENDOR_ID_COCO"
    VENDOR_ID_REGISTER = "DRV_ID_VENDOR_ID"

    def __init__(self, motion_controller: "MotionController") -> None:
        Homing.__init__(self, motion_controller)
        Feedbacks.__init__(self, motion_controller)
        self.mc = motion_controller
        self.logger = ingenialogger.get_logger(__name__)

    def release_brake(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> None:
        """Override the brake status to released in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. 1 by default.

        """
        self.mc.communication.set_register(
            self.BRAKE_OVERRIDE_REGISTER, self.BrakeOverride.RELEASE_BRAKE, servo=servo, axis=axis
        )

    def enable_brake(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> None:
        """Override the brake status of the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. 1 by default.

        """
        self.mc.communication.set_register(
            self.BRAKE_OVERRIDE_REGISTER, self.BrakeOverride.ENABLE_BRAKE, servo=servo, axis=axis
        )

    def disable_brake_override(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> None:
        """Disable the brake override of the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. 1 by default.

        """
        self.mc.communication.set_register(
            self.BRAKE_OVERRIDE_REGISTER,
            self.BrakeOverride.OVERRIDE_DISABLED,
            servo=servo,
            axis=axis,
        )

    def default_brake(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> None:
        """Disable the brake override of the target servo and axis, as
        :func:`disable_brake_override`.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. 1 by default.

        """
        self.disable_brake_override(servo, axis)

    def check_configuration(
        self, config_path: str, axis: Optional[int] = None, servo: str = DEFAULT_SERVO
    ) -> None:
        """Check if the drive is configured in the same way as the given configuration file.
        Compares the value of each register in the given file with the corresponding value in the
        drive.

        Args:
            config_path : config file path to check.
            axis : target axis to load configuration.
                If ``None`` function loads all axis. ``None`` by default.
            servo : servo alias to reference it. ``default`` by default.

        Raises:
            FileNotFoundError: If configuration file does not exist.
            ValueError: If a configuration file from a subnode different from 0
                is attempted to be loaded to subnode 0.
            ValueError: If an invalid subnode is provided.
            ILConfigurationError: If the configuration file differs from the drive state.

        """
        if not path.isfile(config_path):
            raise FileNotFoundError(f"{config_path} file does not exist!")
        servo_inst = self.mc.servos[servo]
        servo_inst.check_configuration(config_path, subnode=axis)
        self.logger.info(
            "Configuration check successfull %s", config_path, drive=self.mc.servo_name(servo)
        )

    def load_configuration(
        self, config_path: str, axis: Optional[int] = None, servo: str = DEFAULT_SERVO
    ) -> None:
        """Load a configuration file to the target servo.

        Args:
            config_path : config file path to load.
            axis : target axis to load configuration.
                If ``None`` function loads all axis. ``None`` by default.
            servo : servo alias to reference it. ``default`` by default.

        Raises:
            FileNotFoundError: If configuration file does not exist.

        """
        if not path.isfile(config_path):
            raise FileNotFoundError("{} file does not exist!".format(config_path))
        servo_inst = self.mc.servos[servo]
        servo_inst.load_configuration(config_path, subnode=axis)
        self.logger.info(
            "Configuration loaded from %s", config_path, drive=self.mc.servo_name(servo)
        )

    def save_configuration(
        self, output_file: str, axis: Optional[int] = None, servo: str = DEFAULT_SERVO
    ) -> None:
        """Save the servo configuration to a target file.

        Args:
            output_file : servo configuration destination file.
            axis : target axis to load configuration.
                If ``None`` function loads all axis. ``None`` by default.
            servo : servo alias to reference it. ``default`` by default.

        """
        servo_inst = self.mc.servos[servo]
        servo_inst.save_configuration(output_file, subnode=axis)
        self.logger.info("Configuration saved to %s", output_file, drive=self.mc.servo_name(servo))

    def store_configuration(self, axis: Optional[int] = None, servo: str = DEFAULT_SERVO) -> None:
        """Store servo configuration to non-volatile memory.

        Args:
            axis : target axis to load configuration.
                If ``None`` function loads all axis. ``None`` by default.
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)
        drive.store_parameters(axis)
        self.logger.info("Configuration stored", drive=self.mc.servo_name(servo))

    def restore_configuration(self, axis: Optional[int] = None, servo: str = DEFAULT_SERVO) -> None:
        """Restore servo to default configuration.

        Args:
            axis : target axis to load configuration.
                If ``None`` function loads all axis. ``None`` by default.
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)
        drive.restore_parameters(axis)
        self.logger.info("Configuration restored", drive=self.mc.servo_name(servo))

    def set_max_acceleration(
        self, acceleration: float, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Update maximum acceleration register.

        .. warning::
            This function is deprecated. Please use
            "set_max_profile_acceleration" or "set_profiler" instead.

        Args:
            acceleration : maximum acceleration in rev/s^2.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            TypeError: If acceleration is not a float.

        """
        self.logger.warning(
            '"set_max_acceleration" is deprecated. '
            'Please use "set_max_profile_acceleration" or '
            '"set_profiler".'
        )
        self.mc.communication.set_register(
            self.PROFILE_MAX_ACCELERATION_REGISTER, acceleration, servo=servo, axis=axis
        )
        self.logger.debug(
            "Max acceleration set to %s", acceleration, axis=axis, drive=self.mc.servo_name(servo)
        )

    def set_max_profile_acceleration(
        self, acceleration: float, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Update maximum profile acceleration register.

        Args:
            acceleration: maximum profile acceleration in rev/s^2.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
        """
        self.mc.communication.set_register(
            self.PROFILE_MAX_ACCELERATION_REGISTER, acceleration, servo=servo, axis=axis
        )
        self.logger.debug(
            "Max profile acceleration set to %s",
            acceleration,
            axis=axis,
            drive=self.mc.servo_name(servo),
        )

    def set_max_profile_deceleration(
        self, deceleration: float, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Update maximum profile deceleration register.

        Args:
            deceleration: maximum profile deceleration in rev/s^2.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
        """
        self.mc.communication.set_register(
            self.PROFILE_MAX_DECELERATION_REGISTER, deceleration, servo=servo, axis=axis
        )
        self.logger.debug(
            "Max profile deceleration set to %s",
            deceleration,
            axis=axis,
            drive=self.mc.servo_name(servo),
        )

    def set_profiler(
        self,
        acceleration: Optional[float] = None,
        deceleration: Optional[float] = None,
        velocity: Optional[float] = None,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        """Set up the acceleration, deceleration and velocity profilers.
        All of these parameters are optional, meaning the user can set only one
        if desired. However, At least a minimum of one of these parameters
        is mandatory to call this function.

        Args:
            acceleration: maximum acceleration in rev/s^2.
            deceleration: maximum deceleration in rev/s^2.
            velocity: maximum profile velocity in rev/s.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            TypeError: Missing arguments. All the arguments given were None.

        """
        if acceleration is None and deceleration is None and velocity is None:
            raise TypeError("Missing arguments. At least one argument is required.")

        if acceleration is not None:
            self.set_max_profile_acceleration(acceleration, servo=servo, axis=axis)

        if deceleration is not None:
            self.set_max_profile_deceleration(deceleration, servo=servo, axis=axis)

        if velocity is not None:
            self.set_max_profile_velocity(velocity, servo=servo, axis=axis)

    def get_max_velocity(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> float:
        """Get the maximum velocity.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            Max velocity.

        Raises:
            TypeError: If the read value has the wrong type.

        """
        max_velocity = self.mc.communication.get_register(
            self.MAX_VELOCITY_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(max_velocity, float):
            raise TypeError("Max velocity value has to be a float")
        return max_velocity

    def set_max_velocity(
        self, velocity: float, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Update maximum velocity register.

        Args:
            velocity: maximum velocity in rev/s.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            TypeError: If velocity is not a float.

        """
        self.mc.communication.set_register(
            self.MAX_VELOCITY_REGISTER, velocity, servo=servo, axis=axis
        )
        self.logger.debug(
            "Max velocity set to %s", velocity, axis=axis, drive=self.mc.servo_name(servo)
        )

    def set_max_profile_velocity(
        self, velocity: float, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Update maximum profile velocity register.

        Args:
            velocity: maximum profile velocity in rev/s.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
        """
        self.mc.communication.set_register(
            self.PROFILE_MAX_VELOCITY_REGISTER, velocity, servo=servo, axis=axis
        )
        self.logger.debug(
            "Max profile velocity set to %s", velocity, axis=axis, drive=self.mc.servo_name(servo)
        )

    def get_position_and_velocity_loop_rate(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> int:
        """Get position & velocity loop rate frequency.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            Position & velocity loop rate frequency in Hz.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        pos_vel_loop_rate = self.mc.communication.get_register(
            self.POSITION_AND_VELOCITY_LOOP_RATE_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(pos_vel_loop_rate, int):
            raise TypeError("Position and velocity loop has to be an integer")
        return pos_vel_loop_rate

    def get_current_loop_rate(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> int:
        """Get current loop rate frequency.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            Current loop rate frequency in Hz.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        current_loop = self.mc.communication.get_register(
            self.CURRENT_LOOP_RATE_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(current_loop, int):
            raise TypeError("Current loop value has to be an integer")
        return current_loop

    def get_power_stage_frequency(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS, raw: bool = False
    ) -> int:
        """Get Power stage frequency register.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.
            raw : if ``False`` return frequency in Hz, if ``True``
                return raw register value. ``False`` by default.

        Returns:
            Frequency in Hz if raw is ``False``, else, raw register value.

        Raises:
            ValueError: If power stage frequency selection register has an
                invalid value.
            TypeError: If some read value has a wrong type.

        """
        pow_stg_freq = self.mc.communication.get_register(
            self.POWER_STAGE_FREQUENCY_SELECTION_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(pow_stg_freq, int):
            raise TypeError("Power stage frequency value has to be an integer")
        if raw:
            return pow_stg_freq
        try:
            pow_stg_freq_reg = self.POWER_STAGE_FREQUENCY_REGISTERS[pow_stg_freq]
        except IndexError:
            raise ValueError("Invalid power stage frequency register")
        freq = self.mc.communication.get_register(pow_stg_freq_reg, servo=servo, axis=axis)
        if not isinstance(freq, int):
            raise TypeError("Frequency value has to be an integer")
        return freq

    def get_power_stage_frequency_enum(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> IntEnum:
        """Return Power stage frequency register enum.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            IntEnum: Enum with power stage frequency available values.

        """
        return self.mc.get_register_enum(self.POWER_STAGE_FREQUENCY_SELECTION_REGISTER, servo, axis)

    @MCMetaClass.check_motor_disabled
    def set_power_stage_frequency(
        self, value: int, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Set power stage frequency from enum value.
        See :func: `get_power_stage_frequency_enum`.

        Args:
            value : Enum value to set power stage frequency.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            IMStatusWordError: If motor is enabled.

        """
        self.mc.communication.set_register(
            self.POWER_STAGE_FREQUENCY_SELECTION_REGISTER, value, servo=servo, axis=axis
        )

    def get_status_word(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> int:
        """Return status word register value.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            Status word.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        status_word = self.mc.communication.get_register(self.STATUS_WORD_REGISTER, servo, axis)
        if not isinstance(status_word, int):
            raise TypeError("Power stage frequency value has to be an integer")
        return status_word

    def is_motor_enabled(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> bool:
        """Return motor status.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            ``True`` if motor is enabled, else ``False``.

        """
        status_word = self.mc.configuration.get_status_word(servo=servo, axis=axis)
        return bool(status_word & self.STATUS_WORD_OPERATION_ENABLED_BIT)

    def is_commutation_feedback_aligned(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> bool:
        """Return commutation feedback aligned status.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            ``True`` if commutation feedback is aligned, else ``False``.

        """
        status_word = self.mc.configuration.get_status_word(servo=servo, axis=axis)
        return bool(status_word & self.STATUS_WORD_COMMUTATION_FEEDBACK_ALIGNED_BIT)

    def set_phasing_mode(
        self, phasing_mode: PhasingMode, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Set phasing mode.

        Args:
            phasing_mode : phasing mode.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        """
        self.mc.communication.set_register(self.PHASING_MODE_REGISTER, phasing_mode, servo, axis)

    def get_phasing_mode(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> Union[PhasingMode, int]:
        """
        Get current phasing mode.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            Phasing mode value.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        phasing_mode = self.mc.communication.get_register(self.PHASING_MODE_REGISTER, servo, axis)
        if not isinstance(phasing_mode, int):
            raise TypeError("Phasing mode value has to be an integer")
        try:
            return PhasingMode(phasing_mode)
        except ValueError:
            return phasing_mode

    def set_generator_mode(
        self, mode: GeneratorMode, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """
        Set generator mode.

        Args:
            mode : generator mode value.
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        """
        self.mc.communication.set_register(self.GENERATOR_MODE_REGISTER, mode, servo, axis)

    def set_motor_pair_poles(
        self, pair_poles: int, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """
        Set motor pair poles.

        Args:
            pair_poles : motor pair poles-
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Raises:
            TypeError: If pair poles is not an int.
            ingenialink.exceptions.ILValueError: If pair poles is less than 0.

        """
        self.mc.communication.set_register(
            self.MOTOR_POLE_PAIRS_REGISTER, pair_poles, servo=servo, axis=axis
        )

    def get_motor_pair_poles(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> int:
        """
        Get motor pair poles.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            Pair poles value.

        Raises:
            TypeError: If some read value has a wrong type.
        """
        pair_poles = self.mc.communication.get_register(
            self.MOTOR_POLE_PAIRS_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(pair_poles, int):
            raise TypeError("Pair poles value has to be an integer")
        return pair_poles

    def get_sto_status(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> int:
        """
        Get STO register

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            STO register value.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        sto_status = self.mc.communication.get_register(
            self.STO_STATUS_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(sto_status, int):
            raise TypeError("STO status value has to be an integer")
        return sto_status

    def is_sto1_active(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> int:
        """
        Get STO1 bit from STO register

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            return value of STO1 bit.

        """
        if self.get_sto_status(servo, axis) & self.STO1_ACTIVE_BIT:
            return 1
        else:
            return 0

    def is_sto2_active(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> int:
        """
        Get STO2 bit from STO register

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            return value of STO2 bit.

        """
        if self.get_sto_status(servo, axis) & self.STO2_ACTIVE_BIT:
            return 1
        else:
            return 0

    def check_sto_power_supply(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> int:
        """
        Get power supply bit from STO register

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            return value of power supply bit.

        """
        if self.get_sto_status(servo, axis) & self.STO_SUPPLY_FAULT_BIT:
            return 1
        else:
            return 0

    def check_sto_abnormal_fault(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> int:
        """
        Get abnormal fault bit from STO register

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            return value of abnormal fault bit.

        """
        if self.get_sto_status(servo, axis) & self.STO_ABNORMAL_FAULT_BIT:
            return 1
        else:
            return 0

    def get_sto_report_bit(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> int:
        """
        Get report bit from STO register

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            return value of report bit.

        """
        if self.get_sto_status(servo, axis) & self.STO_REPORT_BIT:
            return 1
        else:
            return 0

    def is_sto_active(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> bool:
        """
        Check if STO is active

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            ``True`` if STO is active, else ``False``.

        """
        return self.get_sto_status(servo, axis) == self.STO_ACTIVE_STATE

    def is_sto_inactive(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> bool:
        """
        Check if STO is inactive

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            ``True`` if STO is inactive, else ``False``.

        """
        return self.get_sto_status(servo, axis) == self.STO_INACTIVE_STATE

    def is_sto_abnormal_latched(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> bool:
        """
        Check if STO is abnormal latched

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            ``True`` if STO is abnormal latched, else ``False``.

        """
        return self.get_sto_status(servo, axis) == self.STO_LATCHED_STATE

    def change_tcp_ip_parameters(
        self, ip_address: str, subnet_mask: str, gateway: str, servo: str = DEFAULT_SERVO
    ) -> None:
        """Change TCP IP parameters and store it.

        Args:
            ip_address : IP Address to be changed.
            subnet_mask : Subnet mask to be changed.
            gateway : Gateway to be changed.
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)
        if not isinstance(drive, EthernetServo):
            raise IMException("TCP IP parameters can only be changed in ethernet servos.")
        drive.change_tcp_ip_parameters(ip_address, subnet_mask, gateway)

    def store_tcp_ip_parameters(self, servo: str = DEFAULT_SERVO) -> None:
        """Store TCP IP parameters to non-volatile memory.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)
        if not isinstance(drive, EthernetServo):
            raise IMException("TCP IP parameters can only be stored in ethernet servos.")
        drive.store_tcp_ip_parameters()

    def restore_tcp_ip_parameters(self, servo: str = DEFAULT_SERVO) -> None:
        """Restore TCP IP parameters to default values.

        Args:
            servo : servo alias to reference it. ``default`` by default.

        """
        drive = self.mc._get_drive(servo)
        if not isinstance(drive, EthernetServo):
            raise IMException("TCP IP parameters can only be restored in ethernet servos.")
        drive.restore_tcp_ip_parameters()

    def get_drive_info_coco_moco(
        self, alias: str
    ) -> Tuple[List[Optional[int]], List[Optional[int]], List[Optional[str]], List[Optional[int]]]:
        """Get product codes, revision numbers, firmware versions and serial numbers from
        COCO and MOCO.

        Args:
            alias: Servo alias.

        Returns:
            Product codes (COCO, MOCO).
            Revision numbers (COCO, MOCO).
            FW versions (COCO, MOCO).
            Serial numbers (COCO, MOCO).

        """
        prod_codes: List[Optional[int]] = [None, None]
        rev_numbers: List[Optional[int]] = [None, None]
        fw_versions: List[Optional[str]] = [None, None]
        serial_number: List[Optional[int]] = [None, None]

        for subnode in [0, 1]:
            # Product codes
            try:
                prod_codes[subnode] = self.get_product_code(alias, subnode)
            except (
                ILError,
                IMException,
            ) as e:
                self.logger.error(e)
            # Revision numbers
            try:
                rev_numbers[subnode] = self.get_revision_number(alias, subnode)
            except (ILError, IMException) as e:
                self.logger.error(e)
            # FW versions
            try:
                fw_versions[subnode] = self.get_fw_version(alias, subnode)
            except (ILError, IMException) as e:
                self.logger.error(e)
            # Serial numbers
            try:
                serial_number[subnode] = self.get_serial_number(alias, subnode)
            except (ILError, IMException) as e:
                self.logger.error(e)

        return prod_codes, rev_numbers, fw_versions, serial_number

    @staticmethod
    def get_subnode_type(subnode: int) -> TYPE_SUBNODES:
        """Get a subnode type depending on the axis number.

        Args:
            subnode: Axis number of the drive.

        Returns:
            Subnode type.

        Raises:
            ValueError: For negative subnode values.
        """
        if subnode < 0:
            raise ValueError("There are no subnodes with negative values")
        return TYPE_SUBNODES.COCO if subnode == 0 else TYPE_SUBNODES.MOCO

    def get_product_code(self, alias: str, subnode: int) -> int:
        """Get the product code of a drive.

        Args:
            alias: Alias of the drive.
            subnode: Axis number of the drive.

        Returns:
            Product code

        Raises:
            TypeError: If some read value has a wrong type.
        """
        product_code_register = self.PRODUCT_ID_REGISTERS[self.get_subnode_type(subnode)]
        product_code_value = self.mc.communication.get_register(
            product_code_register, alias, axis=subnode
        )
        if not isinstance(product_code_value, int):
            raise TypeError("Product code value has to be an integer")
        return product_code_value

    def get_revision_number(self, alias: str, subnode: int) -> int:
        """Get the revision number of a drive.

        Args:
            alias: Alias of the drive.
            subnode: Axis number of the drive.

        Returns:
            Revision number

        Raises:
            TypeError: If some read value has a wrong type.
        """
        revision_number_register = self.REVISION_NUMBER_REGISTERS[self.get_subnode_type(subnode)]
        revision_number_value = self.mc.communication.get_register(
            revision_number_register, alias, axis=subnode
        )
        if not isinstance(revision_number_value, int):
            raise TypeError("Revision number value has to be an integer")
        return revision_number_value

    def get_serial_number(self, alias: str, subnode: int) -> int:
        """Get the serial number of a drive.

        Args:
            alias: Alias of the drive.
            subnode: Axis number of the drive.

        Returns:
            Serial number

        Raises:
            TypeError: If some read value has a wrong type.
        """
        serial_number_register = self.SERIAL_NUMBER_REGISTERS[self.get_subnode_type(subnode)]
        serial_number_value = self.mc.communication.get_register(
            serial_number_register, alias, axis=subnode
        )
        if not isinstance(serial_number_value, int):
            raise TypeError("Serial number value has to be an integer")
        return serial_number_value

    def get_fw_version(self, alias: str, subnode: int) -> str:
        """Get the firmware version of a drive.

        Args:
            alias: Alias of the drive.
            subnode: Axis number of the drive.

        Returns:
            Firmware version.

        Raises:
            TypeError: If some read value has a wrong type.
        """
        fw_register = self.SOFTWARE_VERSION_REGISTERS[self.get_subnode_type(subnode)]
        fw_value = self.mc.communication.get_register(fw_register, alias, axis=subnode)
        if not isinstance(fw_value, str):
            raise TypeError("Firmware value has to be a string")
        return fw_value

    def get_vendor_id(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> int:
        """Get the vendor ID of a drive.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : servo axis. ``1`` by default.

        Returns:
            Vendor ID.

        Raises:
            TypeError: If the read vendor ID has the wrong type.

        """
        if axis == 0:
            register = self.VENDOR_ID_COCO_REGISTER
        else:
            register = self.VENDOR_ID_REGISTER
        vendor_id = self.mc.communication.get_register(register, servo, axis)
        if not isinstance(vendor_id, int):
            raise TypeError(
                f"Wrong {register} value for axis {axis}. Expected int, got {type(vendor_id)}"
            )
        return vendor_id

    def change_baudrate(self, baud_rate: CAN_BAUDRATE, servo: str = DEFAULT_SERVO) -> None:
        """Change a CANopen device's baudrate.

        Args:
            baud_rate: New baud rate value.
            servo : servo alias to reference it. ``default`` by default.

        Raises:
            ValueError: If the servo is not a CANopen device.

        """
        drive = self.mc._get_drive(servo)
        net = self.mc._get_network(servo)
        if not isinstance(net, CanopenNetwork):
            raise ValueError(f"Servo {servo} is not a CANopen device.")
        vendor_id = self.get_vendor_id(servo)
        prod_code = self.get_product_code(servo, subnode=0)
        rev_number = self.get_revision_number(servo, subnode=0)
        serial_number = self.get_serial_number(servo, subnode=0)
        net.change_baudrate(
            int(drive.target), baud_rate, vendor_id, prod_code, rev_number, serial_number
        )

    def change_node_id(self, node_id: int, servo: str = DEFAULT_SERVO) -> None:
        """Change a CANopen device's node ID.

        Args:
            node_id: New node ID.
            servo : servo alias to reference it. ``default`` by default.

        Raises:
            ValueError: If servo is not a CANopen device.

        """
        drive = self.mc._get_drive(servo)
        net = self.mc._get_network(servo)
        if not isinstance(net, CanopenNetwork):
            raise ValueError(f"Servo {servo} is not a CANopen device.")
        vendor_id = self.get_vendor_id(servo)
        prod_code = self.get_product_code(servo, subnode=0)
        rev_number = self.get_revision_number(servo, subnode=0)
        serial_number = self.get_serial_number(servo, subnode=0)
        net.change_node_id(
            int(drive.target), node_id, vendor_id, prod_code, rev_number, serial_number
        )

    def set_velocity_pid(
        self,
        kp: float,
        ki: float = 0,
        kd: float = 0,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        """Set velocity PID values in the target servo and axis.

        Args:
            kp: proportional constant
            ki: integral constant
            kd: derivative constant
            servo: servo alias to reference it. ``default`` by default.
            axis: servo axis. ``1`` by default.

        """
        self.mc.communication.set_register(
            self.VELOCITY_LOOP_KP_REGISTER, kp, servo=servo, axis=axis
        )
        self.mc.communication.set_register(
            self.VELOCITY_LOOP_KI_REGISTER, ki, servo=servo, axis=axis
        )
        self.mc.communication.set_register(
            self.VELOCITY_LOOP_KD_REGISTER, kd, servo=servo, axis=axis
        )

    def set_position_pid(
        self,
        kp: float,
        ki: float = 0,
        kd: float = 0,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        """Set position PID values in the target servo and axis.

        Args:
            kp: proportional constant
            ki: integral constant
            kd: derivative constant
            servo: servo alias to reference it. ``default`` by default.
            axis: servo axis. ``1`` by default.

        """
        self.mc.communication.set_register(
            self.POSITION_LOOP_KP_REGISTER, kp, servo=servo, axis=axis
        )
        self.mc.communication.set_register(
            self.POSITION_LOOP_KI_REGISTER, ki, servo=servo, axis=axis
        )
        self.mc.communication.set_register(
            self.POSITION_LOOP_KD_REGISTER, kd, servo=servo, axis=axis
        )

    def get_rated_current(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> float:
        """Get rated current in the target servo and axis.

        Args:
            servo: servo alias to reference it. ``default`` by default.
            axis: servo axis. ``1`` by default.

        Returns:
            Rated current

        Raises:
            TypeError: If some read value has a wrong type.

        """
        rated_current = self.mc.communication.get_register(
            self.RATED_CURRENT_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(rated_current, float):
            raise TypeError(
                f"Wrong {self.RATED_CURRENT_REGISTER} value for axis {axis}. "
                f"Expected int, got {type(rated_current)}"
            )
        return rated_current

    def set_rated_current(
        self, rated_current: float, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Set rated current in the target servo and axis.

        Args:
            rated_current: target rated current.
            servo: servo alias to reference it. ``default`` by default.
            axis: servo axis. ``1`` by default.

        """
        self.mc.communication.set_register(
            self.RATED_CURRENT_REGISTER, rated_current, servo=servo, axis=axis
        )

    def get_max_current(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> float:
        """Get max current in the target servo and axis.

        Args:
            servo: servo alias to reference it. ``default`` by default.
            axis: servo axis. ``1`` by default.

        Returns:
            Max current

        Raises:
            TypeError: If some read value has a wrong type.

        """
        max_current = self.mc.communication.get_register(
            self.MAX_CURRENT_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(max_current, float):
            raise TypeError(
                f"Wrong {self.MAX_CURRENT_REGISTER} value for axis {axis}. "
                f"Expected int, got {type(max_current)}"
            )
        return max_current

    def get_commutation_mode(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> CommutationMode:
        """Get commutation mode in the target servo and axis.

        Args:
            servo: servo alias to reference it. ``default`` by default.
            axis: servo axis. ``1`` by default.

        Returns:
            Commutation mode.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        commutation_mode = self.mc.communication.get_register(
            self.COMMUTATION_MODE_REGISTER, servo, axis
        )
        if not isinstance(commutation_mode, int):
            raise TypeError("Commutation mode value has to be an integer")

        return CommutationMode(commutation_mode)

    def set_commutation_mode(
        self,
        commutation_mode: CommutationMode,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        """Set commutation mode in the target servo and axis.

        Args:
            rated_current: target commutation mode.
            servo: servo alias to reference it. ``default`` by default.
            axis: servo axis. ``1`` by default.

        """
        self.mc.communication.set_register(
            self.COMMUTATION_MODE_REGISTER, commutation_mode, servo, axis
        )

    def get_bus_voltage(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> float:
        """Get Bus voltage.

        Args:
            servo: servo alias to reference it. ``default`` by default.
            axis: servo axis. ``1`` by default.

        Returns:
            Bus voltage.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        bus_voltage = self.mc.communication.get_register(self.BUS_VOLTAGE_REGISTER, servo, axis)
        if not isinstance(bus_voltage, float):
            raise TypeError("Bus voltage value has to be a float")

        return bus_voltage

    def get_pos_to_vel_ratio(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> float:
        """Get the position-to-velocity ratio.

        Args:
            servo: servo alias to reference it. ``default`` by default.
            axis: servo axis. ``1`` by default.

        Returns:
            Position-to-velocity ratio.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        pos_to_vel_ratio = self.mc.communication.get_register(
            self.POSITION_TO_VELOCITY_RATIO_REGISTER, servo, axis
        )
        if not isinstance(pos_to_vel_ratio, float):
            raise TypeError("Position-to-velocity ratio value has to be a float")

        return pos_to_vel_ratio

    def set_pos_to_vel_ratio(
        self, pos_to_vel_ratio: float, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Set the position-to-velocity ratio.

        Args:
            pos_to_vel_ratio: position-to-velocity ratio.
            servo: servo alias to reference it. ``default`` by default.
            axis: servo axis. ``1`` by default.

        """
        self.mc.communication.set_register(
            self.POSITION_TO_VELOCITY_RATIO_REGISTER, pos_to_vel_ratio, servo, axis
        )

    def configure_filter(
        self,
        signal: FilterSignal,
        number: FilterNumber,
        filter_type: Optional[FilterType] = None,
        frequency: Optional[int] = None,
        q_factor: Optional[float] = None,
        gain: Optional[float] = None,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        """Configure a filter.

        Args:
            signal: Signal to be filtered.
            number: Filter number (1 or 2).
            filter_type: Filter type.
            frequency: Cut-off frequency.
            q_factor: Q factor.
            gain: Filter gain.
            servo: servo alias to reference it. ``default`` by default.
            axis: servo axis. ``1`` by default.
        """
        arguments = [filter_type, frequency, q_factor, gain]
        registers = [
            self.FILTER_TYPE_REGISTER,
            self.FILTER_FREQ_REGISTER,
            self.FILTER_Q_REGISTER,
            self.FILTER_GAIN_REGISTER,
        ]
        for value, reg_template in zip(arguments, registers):
            if value is None:
                continue
            register = reg_template.format(signal.value, number.value)
            self.mc.communication.set_register(register, value, servo, axis)



from typing import TYPE_CHECKING, List


if TYPE_CHECKING:
    from ingeniamotion import MotionController
from ingeniamotion.wizard_tests.feedbacks_tests.feedback_test import Feedbacks
from ingeniamotion.enums import SensorType


class DigitalIncremental1Test(Feedbacks):
    BACKUP_REGISTERS_QEI: List[str] = ["FBK_DIGENC1_POLARITY"]

    FEEDBACK_POLARITY_REGISTER = "FBK_DIGENC1_POLARITY"

    SENSOR_TYPE_FEEDBACK_TEST = SensorType.QEI

    def __init__(self, mc: "MotionController", servo: str, axis: int) -> None:
        super().__init__(mc, servo, axis)
        self.backup_registers_names.extend(self.BACKUP_REGISTERS_QEI)



from typing import TYPE_CHECKING, List


if TYPE_CHECKING:
    from ingeniamotion import MotionController
from ingeniamotion.wizard_tests.feedbacks_tests.feedback_test import Feedbacks
from ingeniamotion.wizard_tests.base_test import BaseTest
from ingeniamotion.enums import SensorType


class AbsoluteEncoder2Test(Feedbacks):
    BACKUP_REGISTERS_BISSC2: List[str] = ["FBK_BISS2_POS_POLARITY"]

    FEEDBACK_POLARITY_REGISTER = "FBK_BISS2_POS_POLARITY"

    SENSOR_TYPE_FEEDBACK_TEST = SensorType.BISSC2

    def __init__(self, mc: "MotionController", servo: str, axis: int):
        super().__init__(mc, servo, axis)
        self.backup_registers_names.extend(self.BACKUP_REGISTERS_BISSC2)

    @BaseTest.stoppable
    def feedback_setting(self) -> None:
        super().feedback_setting()
        self.mc.configuration.set_auxiliar_feedback(
            SensorType.ABS1, servo=self.servo, axis=self.axis
        )



import pytest

from ingeniamotion.wizard_tests.feedbacks_tests.absolute_encoder1_test import (
    AbsoluteEncoder1Test,
)
from ingeniamotion.wizard_tests.feedbacks_tests.absolute_encoder2_test import (
    AbsoluteEncoder2Test,
)
from ingeniamotion.wizard_tests.feedbacks_tests.digital_hall_test import DigitalHallTest
from ingeniamotion.wizard_tests.feedbacks_tests.digital_incremental1_test import (
    DigitalIncremental1Test,
)
from ingeniamotion.wizard_tests.feedbacks_tests.digital_incremental2_test import (
    DigitalIncremental2Test,
)
from ingeniamotion.wizard_tests.feedbacks_tests.secondary_ssi_test import (
    SecondarySSITest,
)


@pytest.mark.virtual
@pytest.mark.parametrize(
    "feedback_test_type, expected_total_mandatory",
    [
        (DigitalIncremental1Test, 25),
        (DigitalIncremental2Test, 25),
        (AbsoluteEncoder1Test, 25),
        (AbsoluteEncoder2Test, 25),
        (SecondarySSITest, 24),
        (DigitalHallTest, 27),
    ],
)
def test_feedback_test_initialization(
    motion_controller, feedback_test_type, expected_total_mandatory
):
    expected_total_optional = 5
    expected_total_backup_registers = expected_total_mandatory + expected_total_optional

    mc, alias = motion_controller
    axis = 1
    feedback_test = feedback_test_type(mc, alias, axis)

    feedback_test = feedback_test_type(mc, alias, 1)
    total_mandatory = len(feedback_test.backup_registers_names)
    total_optional = len(feedback_test.optional_backup_registers_names)
    total_backup_register = total_mandatory + total_optional

    assert total_mandatory == expected_total_mandatory
    assert total_optional == expected_total_optional
    assert total_backup_register == expected_total_backup_registers


@pytest.mark.virtual
@pytest.mark.parametrize(
    "feedback_test_type",
    [
        DigitalIncremental1Test,
        DigitalIncremental2Test,
        AbsoluteEncoder1Test,
        AbsoluteEncoder2Test,
        SecondarySSITest,
        DigitalHallTest,
    ],
)
def test_save_backup_registers(motion_controller, feedback_test_type):
    mc, alias = motion_controller
    axis = 1
    feedback_test = feedback_test_type(mc, alias, axis)
    mandatory_backup_registers = feedback_test.backup_registers_names

    expected_total_mandatory = len(mandatory_backup_registers)
    expected_total_optional = 3
    expected_total_backup_registers = expected_total_mandatory + expected_total_optional

    feedback_test.save_backup_registers()
    saved_backup_registers = feedback_test.backup_registers[axis]
    total_backup_registers = len(saved_backup_registers)

    assert total_backup_registers == expected_total_backup_registers
    for expected_register in feedback_test.backup_registers_names:
        assert expected_register in saved_backup_registers
    assert "COMMU_ANGLE_INTEGRITY1_OPTION" not in saved_backup_registers
    assert "COMMU_ANGLE_INTEGRITY2_OPTION" not in saved_backup_registers
    assert "PROF_POS_OPTION_CODE" in saved_backup_registers
    assert "CL_POS_REF_MAX_RANGE" in saved_backup_registers
    assert "CL_POS_REF_MIN_RANGE" in saved_backup_registers



import inspect
from functools import wraps
from types import FunctionType
from typing import Any, Callable, Dict, Tuple, Type, TypeVar

from ingeniamotion.exceptions import IMStatusWordError

DEFAULT_SERVO = "default"
DEFAULT_AXIS = 1

T = TypeVar("T")


class MCMetaClass(type):
    """MotionController submodules metaclass to add servo checker for all
    the functions that has an argument named servo.

    This class also have other decorators that can be useful for some
    functions, as motor disabled checker.
    """

    SERVO_ARG_NAME = "servo"
    AXIS_ARG_NAME = "axis"

    def __new__(
        mcs: Type["MCMetaClass"], name: str, bases: Tuple[type, ...], local: Dict[str, Any]
    ) -> "MCMetaClass":
        """If a function has argument named servo,
        decorates it with check_servo decorator.
        """
        for attr in local:
            value = local[attr]
            if (
                callable(value)
                and isinstance(value, FunctionType)
                and mcs.SERVO_ARG_NAME in inspect.getfullargspec(value).args
            ):
                local[attr] = mcs.check_servo(value)
        return type.__new__(mcs, name, bases, local)

    @classmethod
    def check_servo(mcs, func: Callable[..., T]) -> Callable[..., T]:
        """Decorator to check if the servo is connected.
        If servo is not connected raises an exception.
        """

        @wraps(func)
        def wrapper(self, *args, **kwargs):  # type: ignore
            mc = self.mc
            func_args = inspect.getfullargspec(func).args
            servo_index = func_args.index(mcs.SERVO_ARG_NAME)
            if len(args) < servo_index:
                servo = kwargs.get(mcs.SERVO_ARG_NAME, DEFAULT_SERVO)
            else:
                servo = args[servo_index - 1]
            if servo not in mc.servos:
                raise KeyError("Servo '{}' is not connected".format(servo))
            return func(self, *args, **kwargs)

        return wrapper

    @classmethod
    def check_motor_disabled(mcs, func: Callable[..., T]) -> Callable[..., T]:
        """Decorator to check if motor is disabled.
        If motor is enabled raises an exception.
        """

        @wraps(func)
        def wrapper(self, *args, **kwargs):  # type: ignore
            mc = self.mc
            func_args = inspect.getfullargspec(func).args
            servo_index = func_args.index(mcs.SERVO_ARG_NAME)
            axis_index = func_args.index(mcs.AXIS_ARG_NAME)
            if len(args) < servo_index:
                servo = kwargs.get(mcs.SERVO_ARG_NAME, DEFAULT_SERVO)
            else:
                servo = args[servo_index - 1]
            if len(args) < axis_index:
                axis = kwargs.get(mcs.AXIS_ARG_NAME, DEFAULT_AXIS)
            else:
                axis = args[axis_index - 1]
            if mc.configuration.is_motor_enabled(servo=servo, axis=axis):
                raise IMStatusWordError("Motor is enabled")
            return func(self, *args, **kwargs)

        return wrapper



from enum import IntEnum
from typing import Dict, Optional, Union

import ingenialogger

from ingeniamotion.enums import SeverityLevel, OperationMode
from ingeniamotion.exceptions import IMRegisterNotExist
from ingeniamotion.metaclass import DEFAULT_SERVO, DEFAULT_AXIS
from ingeniamotion.motion_controller import MotionController
from ingeniamotion.wizard_tests import stoppable
from ingeniamotion.wizard_tests.base_test import BaseTest


class BrakeRegKey(IntEnum):
    """Brake Register Keys for dictionaries"""

    FEEDBACK_SOURCE = 0
    CONTROL_MODE = 1


class ResultBrakeType(IntEnum):
    """Type of result once a brake tuning is stopped or failed"""

    SUCCESS = 0
    FAIL_FEEDBACK_SOURCE = 1
    FAIL_CURRENT_MODE = 2
    FAIL_DICTIONARY = 3


class BrakeTune(BaseTest):
    """A class to perform a brake tuning. It enables and disables a brake through enabling/disabling the motor.

    Args:
        mc: Motion Controller.
        enable_disable_motor_period: Period of time in seconds between a motor starts to be enabled,
            and it finishes to be disabled.
        servo: servo alias to reference it. ``default`` by default.
        axis: axis that will run the test. 1 by default.

    """

    BRAKE_CURRENT_FEEDBACK_SOURCE = "MOT_BRAKE_CUR_FBK"
    BRAKE_CONTROL_MODE = "MOT_BRAKE_CONTROL_MODE"

    BACKUP_REGISTERS = [
        "MOT_BRAKE_OVERRIDE",
        "DRV_OP_CMD",
        "CL_VOL_Q_SET_POINT",
        "CL_VOL_D_SET_POINT",
    ]

    def __init__(
        self,
        mc: MotionController,
        enable_disable_motor_period: float = 1.0,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        super().__init__()
        self.mc = mc
        self.servo = servo
        self.axis = axis
        self.logger = ingenialogger.get_logger(__name__, axis=axis, drive=mc.servo_name(servo))
        self.backup_registers_names = self.BACKUP_REGISTERS
        self.__enable_disable_motor_period = enable_disable_motor_period

    def setup(self) -> None:
        # Make sure Brake override is disabled
        self.mc.configuration.disable_brake_override(servo=self.servo, axis=self.axis)
        # Make sure the operation mode is set as Voltage mode
        self.mc.motion.set_operation_mode(OperationMode.VOLTAGE, servo=self.servo, axis=self.axis)
        # Make sure quadrature voltage is set to 0v
        self.mc.motion.set_voltage_quadrature(0, servo=self.servo, axis=self.axis)
        # Make sure direct voltage is set to 0v
        self.mc.motion.set_voltage_direct(0, servo=self.servo, axis=self.axis)

    @BaseTest.stoppable
    def loop(self) -> Optional[ResultBrakeType]:
        try:
            reg_values = self.__update_brake_registers_values()
        except IMRegisterNotExist:
            return ResultBrakeType.FAIL_DICTIONARY
        # Register values to avoid meanwhile the test is being executed
        no_brake_current_feedback_source = 0
        brake_is_current_mode = 1
        # Start the test
        try:
            while (
                reg_values[BrakeRegKey.FEEDBACK_SOURCE] != no_brake_current_feedback_source
            ) and (reg_values[BrakeRegKey.CONTROL_MODE] == brake_is_current_mode):
                # Motor enable
                self.mc.motion.motor_enable(servo=self.servo, axis=self.axis)
                self.stoppable_sleep(self.__enable_disable_motor_period / 2)
                # Motor disable
                self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)
                self.stoppable_sleep(self.__enable_disable_motor_period / 2)

                reg_values = self.__update_brake_registers_values()
        except stoppable.StopException:
            self.logger.info("Test has been stopped")
            return ResultBrakeType.SUCCESS

        if reg_values[BrakeRegKey.FEEDBACK_SOURCE] == no_brake_current_feedback_source:
            return ResultBrakeType.FAIL_FEEDBACK_SOURCE
        elif reg_values[BrakeRegKey.CONTROL_MODE] != brake_is_current_mode:
            return ResultBrakeType.FAIL_CURRENT_MODE
        else:
            return None

    def __update_brake_registers_values(self) -> Dict[BrakeRegKey, Union[int, float, str]]:
        brake_registers_updated = {}
        updated_brake_current_feedback_source = self.mc.communication.get_register(
            self.BRAKE_CURRENT_FEEDBACK_SOURCE, servo=self.servo, axis=self.axis
        )
        brake_registers_updated[BrakeRegKey.FEEDBACK_SOURCE] = updated_brake_current_feedback_source
        updated_brake_control_mode = self.mc.communication.get_register(
            self.BRAKE_CONTROL_MODE, servo=self.servo, axis=self.axis
        )
        brake_registers_updated[BrakeRegKey.CONTROL_MODE] = updated_brake_control_mode
        return brake_registers_updated

    def teardown(self) -> None:
        self.logger.info("Disabling brake")
        self.mc.motion.motor_disable(servo=self.servo, axis=self.axis)

    def get_result_severity(self, output: ResultBrakeType) -> SeverityLevel:
        severity_options = {
            ResultBrakeType.SUCCESS: SeverityLevel.SUCCESS,
            ResultBrakeType.FAIL_FEEDBACK_SOURCE: SeverityLevel.FAIL,
            ResultBrakeType.FAIL_CURRENT_MODE: SeverityLevel.FAIL,
            ResultBrakeType.FAIL_DICTIONARY: SeverityLevel.FAIL,
        }
        return severity_options[output]

    def get_result_msg(self, output: ResultBrakeType) -> str:
        message_options = {
            ResultBrakeType.SUCCESS: "Brake tune is stopped properly",
            ResultBrakeType.FAIL_FEEDBACK_SOURCE: "A brake current feedback source is not set",
            ResultBrakeType.FAIL_CURRENT_MODE: "The brake is not in current mode",
            ResultBrakeType.FAIL_DICTIONARY: "Brake current control mode is not implemented in the drive",
        }
        return message_options[output]



from typing import TYPE_CHECKING, Optional, Union

import ingenialogger

from ingeniamotion.enums import SensorType, SensorCategory, FeedbackPolarity

if TYPE_CHECKING:
    from ingeniamotion.motion_controller import MotionController
from ingeniamotion.metaclass import MCMetaClass, DEFAULT_AXIS, DEFAULT_SERVO


class Feedbacks(metaclass=MCMetaClass):
    """Feedbacks Wizard Class description."""

    __feedback_type_dict = {
        SensorType.ABS1: SensorCategory.ABSOLUTE,
        SensorType.QEI: SensorCategory.INCREMENTAL,
        SensorType.HALLS: SensorCategory.ABSOLUTE,
        SensorType.SSI2: SensorCategory.ABSOLUTE,
        SensorType.BISSC2: SensorCategory.ABSOLUTE,
        SensorType.QEI2: SensorCategory.INCREMENTAL,
        SensorType.INTGEN: SensorCategory.ABSOLUTE,
    }

    __feedback_polarity_register_dict = {
        SensorType.ABS1: "FBK_BISS1_SSI1_POS_POLARITY",
        SensorType.QEI: "FBK_DIGENC1_POLARITY",
        SensorType.HALLS: "FBK_DIGHALL_POLARITY",
        SensorType.SSI2: "FBK_SSI2_POS_POLARITY",
        SensorType.BISSC2: "FBK_BISS2_POS_POLARITY",
        SensorType.QEI2: "FBK_DIGENC2_POLARITY",
    }

    COMMUTATION_FEEDBACK_REGISTER = "COMMU_ANGLE_SENSOR"
    REFERENCE_FEEDBACK_REGISTER = "COMMU_ANGLE_REF_SENSOR"
    VELOCITY_FEEDBACK_REGISTER = "CL_VEL_FBK_SENSOR"
    POSITION_FEEDBACK_REGISTER = "CL_POS_FBK_SENSOR"
    AUXILIAR_FEEDBACK_REGISTER = "CL_AUX_FBK_SENSOR"

    def __init__(self, motion_controller: "MotionController") -> None:
        self.mc = motion_controller
        self.logger = ingenialogger.get_logger(__name__)
        self.feedback_resolution_functions = {
            SensorType.ABS1: self.get_absolute_encoder_1_resolution,
            SensorType.QEI: self.get_incremental_encoder_1_resolution,
            SensorType.HALLS: self.get_digital_halls_resolution,
            SensorType.SSI2: self.get_secondary_ssi_resolution,
            SensorType.BISSC2: self.get_absolute_encoder_2_resolution,
            SensorType.QEI2: self.get_incremental_encoder_2_resolution,
            SensorType.INTGEN: self.__no_feedback_resolution,
        }

    # Commutation feedback
    def get_commutation_feedback(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> SensorType:
        """Reads commutation feedbacks value in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Type of feedback configured.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        commutation_feedback = self.mc.communication.get_register(
            self.COMMUTATION_FEEDBACK_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(commutation_feedback, int):
            raise TypeError("Commutation feedback value has to be an integer")
        return SensorType(commutation_feedback)

    @MCMetaClass.check_motor_disabled
    def set_commutation_feedback(
        self, feedback: SensorType, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Writes commutation feedbacks value in the target servo and axis.

        Args:
            feedback : feedback sensor number
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Raises:
            IMStatusWordError: If motor is enabled.
        """
        self.mc.communication.set_register(
            self.COMMUTATION_FEEDBACK_REGISTER, feedback, servo=servo, axis=axis
        )

    def get_commutation_feedback_category(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> SensorCategory:
        """Reads commutation feedbacks type {ABSOLUTE or INCREMENTAL}
        in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Category {ABSOLUTE, INCREMENTAL} of the selected feedback.
        """
        commutation_feedback = self.get_commutation_feedback(servo, axis)
        return self.__feedback_type_dict[commutation_feedback]

    def get_commutation_feedback_resolution(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> int:
        """Reads commutation feedbacks resolution in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Resolution of the selected feedback.
        """
        sensor_type = self.get_commutation_feedback(servo, axis)
        return self.feedback_resolution_functions[sensor_type](servo, axis)

    # Reference feedback
    def get_reference_feedback(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> SensorType:
        """Reads reference feedbacks value in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Type of feedback configured

        Raises:
            TypeError: If some read value has a wrong type.

        """
        reference_feedback = self.mc.communication.get_register(
            self.REFERENCE_FEEDBACK_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(reference_feedback, int):
            raise TypeError("Reference feedback has to be an integer")
        return SensorType(reference_feedback)

    @MCMetaClass.check_motor_disabled
    def set_reference_feedback(
        self, feedback: SensorType, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Writes reference feedbacks value in the target servo and axis.

        Args:
            feedback : feedback sensor number
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Raises:
            IMStatusWordError: If motor is enabled.
        """
        self.mc.communication.set_register(
            self.REFERENCE_FEEDBACK_REGISTER, feedback, servo=servo, axis=axis
        )

    def get_reference_feedback_category(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> SensorCategory:
        """Reads reference feedbacks type {ABSOLUTE or INCREMENTAL}
        in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Category {ABSOLUTE, INCREMENTAL} of the selected feedback.
        """
        reference_feedback = self.get_reference_feedback(servo, axis)
        return self.__feedback_type_dict[reference_feedback]

    def get_reference_feedback_resolution(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> int:
        """Reads reference feedbacks resolution in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Resolution of the selected feedback.
        """
        sensor_type = self.get_reference_feedback(servo, axis)
        return self.feedback_resolution_functions[sensor_type](servo, axis)

    # Velocity feedback
    def get_velocity_feedback(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> SensorType:
        """Reads velocity feedbacks value in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Type of feedback configured

        Raises:
            TypeError: If some read value has a wrong type.

        """
        velocity_feedback = self.mc.communication.get_register(
            self.VELOCITY_FEEDBACK_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(velocity_feedback, int):
            raise TypeError("Velocity feedback has to be an integer")
        return SensorType(velocity_feedback)

    @MCMetaClass.check_motor_disabled
    def set_velocity_feedback(
        self, feedback: SensorType, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Writes velocity feedbacks value in the target servo and axis.

        Args:
            feedback : feedback sensor number
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Raises:
            IMStatusWordError: If motor is enabled.
        """
        self.mc.communication.set_register(
            self.VELOCITY_FEEDBACK_REGISTER, feedback, servo=servo, axis=axis
        )

    def get_velocity_feedback_category(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> SensorCategory:
        """Reads velocity feedbacks type {ABSOLUTE or INCREMENTAL}
        in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Category {ABSOLUTE, INCREMENTAL} of the selected feedback.
        """
        velocity_feedback = self.get_velocity_feedback(servo, axis)
        return self.__feedback_type_dict[velocity_feedback]

    def get_velocity_feedback_resolution(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> int:
        """Reads velocity feedbacks resolution in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Resolution of the selected feedback.
        """
        sensor_type = self.get_velocity_feedback(servo, axis)
        return self.feedback_resolution_functions[sensor_type](servo, axis)

    # Position feedback
    def get_position_feedback(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> SensorType:
        """Reads position feedbacks value in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Type of feedback configured.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        position_feedback = self.mc.communication.get_register(
            self.POSITION_FEEDBACK_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(position_feedback, int):
            raise TypeError("Position feedback has to be an integer")
        return SensorType(position_feedback)

    @MCMetaClass.check_motor_disabled
    def set_position_feedback(
        self, feedback: SensorType, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Writes position feedbacks value in the target servo and axis.

        Args:
            feedback : feedback sensor number
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Raises:
            IMStatusWordError: If motor is enabled.
        """
        self.mc.communication.set_register(
            self.POSITION_FEEDBACK_REGISTER, feedback, servo=servo, axis=axis
        )

    def get_position_feedback_category(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> SensorCategory:
        """Reads position feedbacks type {ABSOLUTE or INCREMENTAL}
        in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Category {ABSOLUTE, INCREMENTAL} of the selected feedback.
        """
        position_feedback = self.get_position_feedback(servo, axis)
        return self.__feedback_type_dict[position_feedback]

    def get_position_feedback_resolution(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> int:
        """Reads position feedbacks resolution in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Resolution of the selected feedback.
        """
        sensor_type = self.get_position_feedback(servo, axis)
        return self.feedback_resolution_functions[sensor_type](servo, axis)

    # Auxiliar feedback
    def get_auxiliar_feedback(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> SensorType:
        """Reads auxiliar feedbacks value in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Type of feedback configured

        Raises:
            TypeError: If some read value has a wrong type.

        """
        auxiliar_feedback = self.mc.communication.get_register(
            self.AUXILIAR_FEEDBACK_REGISTER, servo=servo, axis=axis
        )
        if not isinstance(auxiliar_feedback, int):
            raise TypeError("Auxiliar feedback has to be an integer")
        return SensorType(auxiliar_feedback)

    @MCMetaClass.check_motor_disabled
    def set_auxiliar_feedback(
        self, feedback: SensorType, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> None:
        """Writes auxiliar feedbacks value in the target servo and axis.

        Args:
            feedback : feedback sensor number
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Raises:
            IMStatusWordError: If motor is enabled.
        """
        self.mc.communication.set_register(
            self.AUXILIAR_FEEDBACK_REGISTER, feedback, servo=servo, axis=axis
        )

    def get_auxiliar_feedback_category(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> SensorCategory:
        """Reads auxiliar feedbacks type {ABSOLUTE or INCREMENTAL}
        in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Category {ABSOLUTE, INCREMENTAL} of the selected feedback.
        """
        auxiliar_feedback = self.get_auxiliar_feedback(servo, axis)
        return self.__feedback_type_dict[auxiliar_feedback]

    def get_auxiliar_feedback_resolution(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> Optional[int]:
        """Reads auxiliar feedbacks resolution in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Resolution of the selected feedback.
        """
        sensor_type = self.get_auxiliar_feedback(servo, axis)
        return self.feedback_resolution_functions[sensor_type](servo, axis)

    def get_absolute_encoder_1_resolution(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> int:
        """Reads ABS1 encoder resolution in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Resolution of ABS1 encoder.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        single_turn_bits = self.mc.communication.get_register(
            "FBK_BISS1_SSI1_POS_ST_BITS", servo=servo, axis=axis
        )
        if not isinstance(single_turn_bits, int):
            raise TypeError("Single-turn bits has to be an integer")
        resolution = 2**single_turn_bits
        if not isinstance(resolution, int):
            raise TypeError("Resolution value has to be an integer")
        return resolution

    def get_incremental_encoder_1_resolution(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> int:
        """Reads incremental encoder 1 resolution in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Resolution of incremental encoder 1.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        resolution = self.mc.communication.get_register(
            "FBK_DIGENC1_RESOLUTION", servo=servo, axis=axis
        )
        if not isinstance(resolution, int):
            raise TypeError("Resolution value has to be an integer")
        return resolution

    def get_digital_halls_resolution(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> int:
        """Reads digital halls pole pairs in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Resolution of digital halls encoder.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        pair_poles = self.mc.communication.get_register(
            "FBK_DIGHALL_PAIRPOLES", servo=servo, axis=axis
        )
        resolution = 6 * pair_poles
        if not isinstance(resolution, int):
            raise TypeError("Resolution value has to be an integer")
        return resolution

    def get_secondary_ssi_resolution(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> int:
        """Reads secondary SSI encoder resolution in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Resolution of secondary SSI encoder.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        secondary_single_turn_bits = self.mc.communication.get_register(
            "FBK_SSI2_POS_ST_BITS", servo=servo, axis=axis
        )
        if not isinstance(secondary_single_turn_bits, int):
            raise TypeError("Resolution value has to be an integer")
        resolution = int(2**secondary_single_turn_bits)
        return resolution

    def get_absolute_encoder_2_resolution(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> int:
        """Reads ABS2 encoder resolution in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Resolution of ABS2 encoder.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        serial_slave_1_single_turn_bits = self.mc.communication.get_register(
            "FBK_BISS2_POS_ST_BITS", servo=servo, axis=axis
        )
        if not isinstance(serial_slave_1_single_turn_bits, int):
            raise TypeError("Single-turn bits has to be an integer")
        resolution = int(2**serial_slave_1_single_turn_bits)
        return resolution

    def get_incremental_encoder_2_resolution(
        self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> int:
        """Reads incremental encoder 2 resolution in the target servo and axis.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Resolution of incremental encoder 2 encoder.

        Raises:
            TypeError: If some read value has a wrong type.

        """
        resolution = self.mc.communication.get_register(
            "FBK_DIGENC2_RESOLUTION", servo=servo, axis=axis
        )
        if not isinstance(resolution, int):
            raise TypeError("Resolution value has to be an integer")
        return resolution

    def __no_feedback_resolution(self, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS) -> int:
        """Used for feedbacks that has no resolution.

        Args:
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Raises:
            ValueError: Selected feedback does not have resolution
        """
        raise ValueError("Selected feedback does not have resolution")

    def get_feedback_resolution(
        self, feedback: SensorType, servo: str = DEFAULT_SERVO, axis: int = DEFAULT_AXIS
    ) -> int:
        """Reads target feedback resolution in the target servo and axis.

        Args:
            feedback : target feedback.
            servo : servo alias to reference it. ``default`` by default.
            axis : axis that will run the test. ``1`` by default.

        Returns:
            Resolution of target feedback.
        """
        return self.feedback_resolution_functions[feedback](servo, axis)

    def get_feedback_polarity_register_uid(self, feedback: SensorType) -> str:
        """Returns feedback polarity register UID

        Args:
           feedback: target feedback sensor.

        Returns:
            Register UID

        """
        polarity_register = self.__feedback_polarity_register_dict.get(feedback)
        if polarity_register is None:
            raise NotImplementedError(f"Sensor {feedback.name} polarity is not implemented")
        return polarity_register

    def set_feedback_polarity(
        self,
        polarity: FeedbackPolarity,
        feedback: SensorType,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> None:
        """Set target feedback polarity in the target servo and axis.

        Args:
            polarity: target polarity.
            feedback: target feedback.
            servo: servo alias to reference it. ``default`` by default.
            axis: axis that will run the test. ``1`` by default.

        """
        polarity_register = self.get_feedback_polarity_register_uid(feedback)
        self.mc.communication.set_register(polarity_register, polarity, servo=servo, axis=axis)
        self.logger.debug(
            f"Feedback {feedback.name} polarity set to {polarity.name}",
            axis=axis,
            drive=self.mc.servo_name(servo),
        )

    def get_feedback_polarity(
        self,
        feedback: SensorType,
        servo: str = DEFAULT_SERVO,
        axis: int = DEFAULT_AXIS,
    ) -> Union[int, FeedbackPolarity]:
        """Get target feedback polarity of the target servo and axis.

        Args:
            feedback: target feedback.
            servo: servo alias to reference it. ``default`` by default.
            axis: axis that will run the test. ``1`` by default.

        Returns:
            Feedback polarity

        Raises:
            TypeError: If some read value has a wrong type.

        """
        polarity_register = self.get_feedback_polarity_register_uid(feedback)
        raw_polarity = self.mc.communication.get_register(polarity_register, servo=servo, axis=axis)
        if not isinstance(raw_polarity, int):
            raise TypeError("Polarity value has to be an integer")
        try:
            return FeedbackPolarity(raw_polarity)
        except ValueError:
            return raw_polarity



from typing import TYPE_CHECKING, List


if TYPE_CHECKING:
    from ingeniamotion import MotionController
from ingeniamotion.wizard_tests.feedbacks_tests.feedback_test import Feedbacks
from ingeniamotion.enums import SensorType


class SecondarySSITest(Feedbacks):
    BACKUP_REGISTERS_SSI2: List[str] = []

    FEEDBACK_POLARITY_REGISTER = "FBK_SSI2_POS_POLARITY"

    SENSOR_TYPE_FEEDBACK_TEST = SensorType.SSI2

    def __init__(self, mc: "MotionController", servo: str, axis: int):
        super().__init__(mc, servo, axis)
        self.backup_registers_names.extend(self.BACKUP_REGISTERS_SSI2)



import pytest

from ingeniamotion.exceptions import IMStatusWordError
from ingeniamotion import MotionController
from ingeniamotion.enums import OperationMode
from ingeniamotion.capture import Capture
from ingeniamotion.configuration import Configuration
from ingeniamotion.motion import Motion
from ingeniamotion.communication import Communication
from ingeniamotion.drive_tests import DriveTests
from ingeniamotion.errors import Errors
from ingeniamotion.information import Information
from ingeniamotion.metaclass import MCMetaClass


@pytest.mark.virtual
@pytest.mark.smoke
def test_motion_controller():
    mc = MotionController()
    assert isinstance(mc.configuration, Configuration)
    assert isinstance(mc.motion, Motion)
    assert isinstance(mc.capture, Capture)
    assert isinstance(mc.communication, Communication)
    assert isinstance(mc.tests, DriveTests)
    assert isinstance(mc.errors, Errors)
    assert isinstance(mc.info, Information)


@pytest.mark.virtual
@pytest.mark.smoke
def test_servo_name(motion_controller):
    mc, alias = motion_controller
    prod_code = mc.servos[alias].info["product_code"]
    servo_arg = () if alias == "default" else (alias,)
    name = mc.servo_name(*servo_arg)
    assert name == "{} ({})".format(prod_code, alias)


@pytest.mark.virtual
@pytest.mark.smoke
def test_get_register_enum(motion_controller):
    mc, alias = motion_controller
    servo_arg = () if alias == "default" else (alias,)
    operation_mode_enum = mc.get_register_enum("DRV_OP_VALUE", *servo_arg)
    operation_mode_values = [op.value for op in operation_mode_enum]
    checked_ops = 0
    for element in OperationMode:
        if element.value in operation_mode_values:
            test_name = operation_mode_enum(element.value).name.replace("-", " ")
            name = element.name.replace("_", " ")
            assert test_name.upper() == name
            checked_ops += 1

    assert checked_ops > 0


@pytest.mark.virtual
@pytest.mark.smoke
def test_is_alive(motion_controller):
    mc, alias = motion_controller
    assert mc.is_alive(alias)


@pytest.mark.virtual
class TestMetaclass:
    class DummyClass(metaclass=MCMetaClass):
        mc = MotionController()

        def is_motor_enabled(self, servo: str, axis: int):
            return servo == "a" and axis == 1

        def __init__(self):
            self.mc.servos = ["a", "b"]
            self.mc.configuration.is_motor_enabled = self.is_motor_enabled

        @MCMetaClass.check_motor_disabled
        def dummy_func(self, servo: str, axis: int):
            pass

    @pytest.mark.virtual
    @pytest.mark.parametrize(
        "servo, axis, error",
        [
            ("a", 1, IMStatusWordError),
            ("b", 1, None),
            ("c", 1, KeyError),
            ("a", 2, None),
        ],
    )
    def test_check_motor_disabled_keyword_parameters(self, servo, axis, error):
        dummy_class = self.DummyClass()
        if error is None:
            dummy_class.dummy_func(servo=servo, axis=axis)
        else:
            with pytest.raises(error):
                dummy_class.dummy_func(servo=servo, axis=axis)

    @pytest.mark.virtual
    @pytest.mark.parametrize(
        "servo, axis, error",
        [
            ("a", 1, IMStatusWordError),
            ("b", 1, None),
            ("c", 1, KeyError),
            ("a", 2, None),
        ],
    )
    def test_check_motor_disabled_positional_parameters(self, servo, axis, error):
        dummy_class = self.DummyClass()
        if error is None:
            dummy_class.dummy_func(servo, axis)
        else:
            with pytest.raises(error):
                dummy_class.dummy_func(servo, axis)
